# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.table import Cell, Row, Table


@pytest.fixture
def row(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    table = document.body.get_table(name="Example1")
    yield table.get_row(1)


@pytest.fixture
def row_repeats(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    table = document.body.get_table(name="Example1")
    yield table.get_row(0)


def test_on_empty_row():
    row = Row()
    assert row.get_values() == []
    assert row.get_values(complete=True) == []
    assert row.get_values(complete=True, get_type=True) == []
    assert row.get_values((0, 10)) == []
    assert row.get_values(cell_type="all") == []
    assert row.get_values(cell_type="All") == []
    assert row.get_values((2, 3), complete=True) == []


def test_get_values_count(row):
    assert len(row.get_values()) == 7


def test_get_values_coord(row):
    coord = (0, 8)
    assert len(row.get_values(coord)) == 7
    coord = "a1:c2"
    assert len(row.get_values(coord)) == 3
    coord = "a1:a2"
    assert len(row.get_values(coord)) == 1
    coord = "a1:EE2"
    assert len(row.get_values(coord)) == 7
    coord = "D1"
    assert len(row.get_values(coord)) == 0
    coord = "c5:a1"
    assert len(row.get_values(coord)) == 0
    coord = (5, 6)
    assert len(row.get_values(coord)) == 2
    coord = (-5, 6)
    assert len(row.get_values(coord)) == 5
    coord = (0, -1)
    assert len(row.get_values(coord)) == 7
    coord = (0, -2)
    assert len(row.get_values(coord)) == 6
    coord = (-1, -1)
    assert len(row.get_values(coord)) == 1
    coord = (1, 0)
    assert len(row.get_values(coord)) == 0


def test_get_values_cell_type(row):
    values = row.get_values(cell_type="all")
    assert len(values) == 7
    values = row.get_values(cell_type="float")
    assert len(values) == 7
    assert values == [1, 1, 1, 2, 3, 3, 3]
    values = row.get_values(cell_type="percentage")
    assert len(values) == 0
    values = row.get_values(cell_type="string")
    assert len(values) == 0


def test_get_values_cell_type2(row):
    row.append_cell(Cell(value="bob"), clone=False)
    row.append_cell(Cell(value=14, cell_type="percentage"))
    row.append_cell(Cell(value="bob2"), clone=False)
    values = row.get_values(cell_type="all")
    assert len(values) == 7 + 3
    assert values == [1, 1, 1, 2, 3, 3, 3, "bob", 14, "bob2"]
    values = row.get_values(cell_type="float")
    assert len(values) == 7
    assert values == [1, 1, 1, 2, 3, 3, 3]
    values = row.get_values(cell_type="percentage")
    assert len(values) == 1
    assert values == [14]
    values = row.get_values(cell_type="string")
    assert len(values) == 2
    assert values == ["bob", "bob2"]


def test_get_values_cell_type2_with_hole(row):
    row.append_cell(Cell(value="bob"), clone=False)
    row.append_cell(Cell(value=14, cell_type="percentage"))
    row.append_cell(Cell(value="bob2"), clone=False)
    row.set_cell(12, Cell(value="far"), clone=False)
    values = row.get_values(cell_type="all")  # aka all non empty
    assert len(values) == 11
    assert values == [1, 1, 1, 2, 3, 3, 3, "bob", 14, "bob2", "far"]
    values = row.get_values()  # difference when requestion everything
    assert len(values) == 13
    assert values == [1, 1, 1, 2, 3, 3, 3, "bob", 14, "bob2", None, None, "far"]
    values = row.get_values(cell_type="float")
    assert len(values) == 7
    assert values == [1, 1, 1, 2, 3, 3, 3]
    values = row.get_values(cell_type="percentage")
    assert len(values) == 1
    assert values == [14]
    values = row.get_values(cell_type="string")
    assert len(values) == 3
    assert values == ["bob", "bob2", "far"]
    values = row.get_values(":")  # requesting everything
    assert len(values) == 13
    assert values == [1, 1, 1, 2, 3, 3, 3, "bob", 14, "bob2", None, None, "far"]
    values = row.get_values(cell_type="float")
    assert len(values) == 7
    assert values == [1, 1, 1, 2, 3, 3, 3]
    values = row.get_values(cell_type="percentage")
    assert len(values) == 1
    assert values == [14]
    values = row.get_values(cell_type="string")
    assert len(values) == 3
    assert values == ["bob", "bob2", "far"]
    values = row.get_values(":")  # requesting everything
    assert len(values) == 13
    assert values == [1, 1, 1, 2, 3, 3, 3, "bob", 14, "bob2", None, None, "far"]
    values = row.get_values(cell_type="float")
    assert len(values) == 7
    assert values == [1, 1, 1, 2, 3, 3, 3]
    values = row.get_values(cell_type="percentage")
    assert len(values) == 1
    assert values == [14]
    values = row.get_values(cell_type="string")
    assert len(values) == 3
    assert values == ["bob", "bob2", "far"]
    values = row.get_values("")  # requesting everything 2
    assert len(values) == 13
    assert values == [1, 1, 1, 2, 3, 3, 3, "bob", 14, "bob2", None, None, "far"]
    values = row.get_values(cell_type="float")
    assert len(values) == 7
    assert values == [1, 1, 1, 2, 3, 3, 3]
    values = row.get_values(cell_type="percentage")
    assert len(values) == 1
    assert values == [14]
    values = row.get_values(cell_type="string")
    assert len(values) == 3
    assert values == ["bob", "bob2", "far"]


def test_get_values_cell_type2_with_hole_and_get_type(row):
    row.append_cell(Cell(value="bob"), clone=False)
    row.append_cell(Cell(value=14, cell_type="percentage"))
    row.append_cell(Cell(value="bob2"), clone=False)
    row.set_cell(12, Cell(value="far"), clone=False)
    values = row.get_values(cell_type="all", get_type=True)  # aka all non empty
    assert len(values) == 11
    assert values == [
        (1, "float"),
        (1, "float"),
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
        (3, "float"),
        ("bob", "string"),
        (14, "percentage"),
        ("bob2", "string"),
        ("far", "string"),
    ]
    values = row.get_values(get_type=True)  # difference when  everything
    assert len(values) == 13
    assert values == [
        (1, "float"),
        (1, "float"),
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
        (3, "float"),
        ("bob", "string"),
        (14, "percentage"),
        ("bob2", "string"),
        (None, None),
        (None, None),
        ("far", "string"),
    ]
    values = row.get_values(cell_type="float", get_type=True)
    assert len(values) == 7
    assert values == [
        (1, "float"),
        (1, "float"),
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
        (3, "float"),
    ]
    values = row.get_values(cell_type="percentage", get_type=True)
    assert len(values) == 1
    assert values == [(14, "percentage")]
    values = row.get_values(cell_type="string", get_type=True)
    assert len(values) == 3
    assert values == [("bob", "string"), ("bob2", "string"), ("far", "string")]


def test_get_values_cell_type2_complete(row):
    row.append_cell(Cell(value="bob"), clone=False)
    row.append_cell(Cell(value=14, cell_type="percentage"))
    row.append_cell(Cell(value="bob2"), clone=False)
    row.set_cell(12, Cell(value="far"), clone=False)
    values = row.get_values(cell_type="ALL", complete=True)
    assert len(values) == 13
    assert values == [1, 1, 1, 2, 3, 3, 3, "bob", 14, "bob2", None, None, "far"]
    values = row.get_values(cell_type="FLOAT", complete=True)
    assert len(values) == 13
    assert values == [1, 1, 1, 2, 3, 3, 3, None, None, None, None, None, None]
    values = row.get_values(cell_type="percentage", complete=True)
    assert len(values) == 13
    assert values == [
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        14,
        None,
        None,
        None,
        None,
    ]
    values = row.get_values(cell_type="string", complete=True)
    assert len(values) == 13
    assert values == [
        None,
        None,
        None,
        None,
        None,
        None,
        None,
        "bob",
        None,
        "bob2",
        None,
        None,
        "far",
    ]


def test_get_values_cell_type_and_coord(row):
    values = row.get_values(coord=(0, 5), cell_type="all")
    assert len(values) == 6
    values = row.get_values(coord=(0, 5), cell_type="float")
    assert len(values) == 6
    values = row.get_values(coord=(0, 5), cell_type="percentage")
    assert len(values) == 0
    values = row.get_values(coord=(2, 5), cell_type="string")
    assert len(values) == 0


def test_get_values_cell_type_and_coord_strange(row):
    values = row.get_values(coord="A:F", cell_type="all")
    assert len(values) == 6
    values = row.get_values(coord="C:", cell_type="all")
    assert len(values) == 5
    values = row.get_values(coord="A : f", cell_type="float")
    assert len(values) == 6
    values = row.get_values(coord="A:F", cell_type="percentage")
    assert len(values) == 0
    values = row.get_values(coord="C:F", cell_type="string")
    assert len(values) == 0


def test_get_values_cell_type_and_coord_strange_long(row):
    values = row.get_values(coord="A8:F2", cell_type="all")
    assert len(values) == 6
    values = row.get_values(coord="C5:", cell_type="all")
    assert len(values) == 5
    values = row.get_values(coord="C5:99", cell_type="all")
    assert len(values) == 5
    values = row.get_values(coord="A2 : f", cell_type="float")
    assert len(values) == 6
    values = row.get_values(coord="A:F5", cell_type="percentage")
    assert len(values) == 0
    values = row.get_values(coord="C:F4", cell_type="string")
    assert len(values) == 0


def test_get_values_cell_type_and_coord_and_get_type(row):
    values = row.get_values(coord=(0, 5), cell_type="all", get_type=True)
    assert len(values) == 6
    assert values == [
        (1, "float"),
        (1, "float"),
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
    ]
    values = row.get_values(coord=(0, 5), cell_type="float", get_type=True)
    assert len(values) == 6
    assert values == [
        (1, "float"),
        (1, "float"),
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
    ]
    values = row.get_values(coord=(0, 5), cell_type="percentage", get_type=True)
    assert len(values) == 0
    values = row.get_values(coord=(2, 5), cell_type="string", get_type=True)
    assert len(values) == 0


def test_get_values_cell_type_and_coord_and_complete(row):
    values = row.get_values(coord=(0, 5), cell_type="all", complete=True)
    assert len(values) == 6
    assert values == [1, 1, 1, 2, 3, 3]
    values = row.get_values(coord=(0, 5), cell_type="float", complete=True)
    assert len(values) == 6
    assert values == [1, 1, 1, 2, 3, 3]
    values = row.get_values(coord=(0, 5), cell_type="percentage", complete=True)
    assert len(values) == 6
    assert values == [None, None, None, None, None, None]
    values = row.get_values(coord=(2, 5), cell_type="string", complete=True)
    assert len(values) == 4
    assert values == [None, None, None, None]


def test_get_values_cell_type_and_coord2_and_complete(row):
    row.append_cell(Cell(value="bob"), clone=False)
    row.append_cell(Cell(value=14, cell_type="percentage"))
    row.append_cell(Cell(value="bob2"), clone=False)
    row.set_cell(12, Cell(value="far"), clone=False)
    values = row.get_values(coord=(2, 20), cell_type="all", complete=True)
    assert len(values) == 11
    assert values == [1, 2, 3, 3, 3, "bob", 14, "bob2", None, None, "far"]
    values = row.get_values(coord=(2, 11), cell_type="all", complete=True)
    assert len(values) == 10
    assert values == [1, 2, 3, 3, 3, "bob", 14, "bob2", None, None]
    values = row.get_values(coord=(3, 12), cell_type="float", complete=True)
    assert len(values) == 10
    assert values == [2, 3, 3, 3, None, None, None, None, None, None]
    values = row.get_values(coord=(0, 5), cell_type="percentage", complete=True)
    assert len(values) == 6
    assert values == [None, None, None, None, None, None]
    values = row.get_values(coord=(0, 5), cell_type="string", complete=True)
    assert len(values) == 6
    assert values == [None, None, None, None, None, None]
    values = row.get_values(coord=(5, 11), cell_type="percentage", complete=True)
    assert len(values) == 7
    assert values == [None, None, None, 14, None, None, None]
    values = row.get_values(coord=(6, 12), cell_type="string", complete=True)
    assert len(values) == 7
    assert values == [None, "bob", None, "bob2", None, None, "far"]
    values = row.get_values(coord=(8, 20), cell_type="string", complete=True)
    assert len(values) == 5
    assert values == [None, "bob2", None, None, "far"]


def test_get_values_cell_type_and_coord2_and_complete_and_get_type(row):
    row.append_cell(Cell(value="bob"), clone=False)
    row.append_cell(Cell(value=14, cell_type="percentage"))
    row.append_cell(Cell(value="bob2"), clone=False)
    row.set_cell(12, Cell(value="far"), clone=False)
    values = row.get_values(
        coord=(2, 20), cell_type="all", complete=True, get_type=True
    )
    assert len(values) == 11
    assert values == [
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
        (3, "float"),
        ("bob", "string"),
        (14, "percentage"),
        ("bob2", "string"),
        (None, None),
        (None, None),
        ("far", "string"),
    ]
    values = row.get_values(
        coord=(2, 11), cell_type="all", complete=True, get_type=True
    )
    assert len(values) == 10
    assert values == [
        (1, "float"),
        (2, "float"),
        (3, "float"),
        (3, "float"),
        (3, "float"),
        ("bob", "string"),
        (14, "percentage"),
        ("bob2", "string"),
        (None, None),
        (None, None),
    ]
    values = row.get_values(
        coord=(3, 12), cell_type="float", complete=True, get_type=True
    )
    assert len(values) == 10
    assert values == [
        (2, "float"),
        (3, "float"),
        (3, "float"),
        (3, "float"),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
    ]
    values = row.get_values(
        coord=(0, 5), cell_type="percentage", complete=True, get_type=True
    )
    assert len(values) == 6
    assert values == [
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
    ]
    values = row.get_values(
        coord=(0, 5), cell_type="string", complete=True, get_type=True
    )
    assert len(values) == 6
    assert values == [
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
    ]
    values = row.get_values(
        coord=(5, 11), cell_type="percentage", complete=True, get_type=True
    )
    assert len(values) == 7
    assert values == [
        (None, None),
        (None, None),
        (None, None),
        (14, "percentage"),
        (None, None),
        (None, None),
        (None, None),
    ]
    values = row.get_values(
        coord=(6, 12), cell_type="string", complete=True, get_type=True
    )
    assert len(values) == 7
    assert values == [
        (None, None),
        ("bob", "string"),
        (None, None),
        ("bob2", "string"),
        (None, None),
        (None, None),
        ("far", "string"),
    ]
    values = row.get_values(
        coord=(8, 20), cell_type="string", complete=True, get_type=True
    )
    assert len(values) == 5
    assert values == [
        (None, None),
        ("bob2", "string"),
        (None, None),
        (None, None),
        ("far", "string"),
    ]


def test_get_cell_alpha(row):
    cell_5 = row.get_cell("F")
    assert cell_5.get_value() == 3
    assert cell_5.text_content == "3"
    assert cell_5.type == "float"
    assert cell_5.style == "ce5"
    assert cell_5.x == 5
    assert cell_5.y == 1


def test_get_cell_int(row):
    cell_5 = row.get_cell(5)
    assert cell_5.get_value() == 3
    assert cell_5.text_content == "3"
    assert cell_5.type == "float"
    assert cell_5.style == "ce5"


def test_get_cell_coord(row):
    cell = row.get_cell(-1)
    assert cell.get_value() == 3
    cell = row.get_cell(-3)
    assert cell.get_value() == 3
    cell = row.get_cell(-4)
    assert cell.get_value() == 2
    cell = row.get_cell(-5)
    assert cell.get_value() == 1
    cell = row.get_cell(-1 - 7)
    assert cell.get_value() == 3
    cell = row.get_cell(-3 - 56)
    assert cell.get_value() == 3
    cell = row.get_cell(-4 - 560)
    assert cell.get_value() == 2
    cell = row.get_cell(-5 - 7000)
    assert cell.get_value() == 1
    cell = row.get_cell(8)
    assert cell.get_value() is None
    cell = row.get_cell(1000)
    assert cell.get_value() is None


def test_get_cells_coord_none(row):
    cells = row.get_cells((100, 200))
    assert len(cells) == 0


def test_get_cell_coord_none(row):
    cell = row.get_cell(2000)
    assert cell


def test_get_value_coord(row):
    row.append_cell(Cell("Appended"))
    value = row.get_value(-1)
    assert value == "Appended"
    value = row.get_value(-3)
    assert value == 3
    value = row.get_value(-4)
    assert value == 3
    value = row.get_value(-5)
    assert value == 2
    value = row.get_value(-1 - 8)
    assert value == "Appended"
    value = row.get_value(7)
    assert value == "Appended"
    value = row.get_value(8)
    assert value is None
    value = row.get_value(1000)
    assert value is None


def test_get_value_coord_with_get_type(row):
    row.append_cell(Cell("Appended"))
    value = row.get_value(-1, get_type=True)
    assert value == ("Appended", "string")
    value = row.get_value(-3, get_type=True)
    assert value == (3, "float")
    value = row.get_value(-4, get_type=True)
    assert value == (3, "float")
    value = row.get_value(-5, get_type=True)
    assert value == (2, "float")
    value = row.get_value(-1 - 8, get_type=True)
    assert value == ("Appended", "string")
    value = row.get_value(7, get_type=True)
    assert value == ("Appended", "string")
    value = row.get_value(8, get_type=True)
    assert value == (None, None)
    value = row.get_value(1000, get_type=True)
    assert value == (None, None)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
"""Representation of the "meta.xml" part."""

from __future__ import annotations

import json
from datetime import date as dtdate
from datetime import datetime, timedelta
from decimal import Decimal
from operator import itemgetter
from typing import TYPE_CHECKING, Any, cast

from .datatype import DateTime, Duration
from .element import Element
from .meta_auto_reload import MetaAutoReload
from .meta_hyperlink_behaviour import MetaHyperlinkBehaviour
from .meta_template import MetaTemplate
from .meta_user_defined import MetaUserDefined
from .mixin_dc_creator import DcCreatorMixin
from .mixin_dc_date import DcDateMixin
from .utils import is_RFC3066, to_str
from .version import __version__
from .xmlpart import XmlPart

if TYPE_CHECKING:
    from .body import Metadata

GENERATOR = f"odfdo {__version__}"


class Meta(XmlPart, DcCreatorMixin, DcDateMixin):
    """Representation of the "meta.xml" part."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the Meta XML part.

        This represents the `meta.xml` file within an ODF document, containing
        various document metadata.

        Args:
            *args: Positional arguments passed to the parent `XmlPart` constructor.
            **kwargs: Keyword arguments passed to the parent `XmlPart` constructor.
        """
        super().__init__(*args, **kwargs)
        self._generator_modified: bool = False

    def _get_body(self) -> Metadata:
        """Internal method to get the root `office:meta` element.

        Returns:
            Metadata: The `office:meta` element.
        """
        return self.get_element("//office:meta")  # type: ignore[return-value]

    get_meta_body = _get_body

    @property
    def meta_body(self) -> Metadata:
        """Return the body of the "meta.xml" part."""
        return self._get_body()

    @property
    def odf_office_version(self) -> str:
        """Return the "office:version" value of the document."""
        odm = self.get_element("//office:document-meta")
        # "office:version" sould be always present
        if odm:
            return odm.get_attribute_string("office:version") or ""
        return ""  # pragma: nocover

    def get_title(self) -> str | None:
        """Get the title of the document.

        This is not the first heading but the title metadata.

        (Also available as "self.title" property.)

        Returns:
            str | None: The title string, or None if inexistent.
        """
        element = self.get_element("//dc:title")
        if element is None:
            return None
        return element.text

    def set_title(self, title: str) -> None:
        """Set the title of the document.

        This is not the first heading but the title metadata.

        (Also available as "self.title" property.)

        Args:
            title: The title string to set.
        """
        element = self.get_element("//dc:title")
        if element is None:
            element = Element.from_tag("dc:title")
            self.body.append(element)
        element.text = title

    @property
    def title(self) -> str | None:
        """Get or set the title of the document <dc:title>.

        Returns: str (or None if inexistent)
        """
        return self.get_title()

    @title.setter
    def title(self, title: str) -> None:
        return self.set_title(title)

    def get_description(self) -> str | None:
        """Get the description of the document. Also known as comments.

        (Also available as "self.description" property.)

        Returns:
            str | None: The description string, or None if inexistent.
        """
        element = self.get_element("//dc:description")
        if element is None:
            return None
        return element.text

    # As named in OOo
    get_comments = get_description

    def set_description(self, description: str) -> None:
        """Set the description of the document. Also known as comments.

        (Also available as "self.description" property.)

        Args:
            description: The description string to set.
        """
        element = self.get_element("//dc:description")
        if element is None:
            element = Element.from_tag("dc:description")
            self.body.append(element)
        element.text = description

    set_comments = set_description

    @property
    def description(self) -> str | None:
        """Get or set the description of a document <dc:description>.

        Returns: str (or None if inexistent)
        """
        return self.get_description()

    @description.setter
    def description(self, description: str) -> None:
        return self.set_description(description)

    def get_subject(self) -> str | None:
        """Get the subject of the document.

        (Also available as "self.subject" property.)

        Returns:
            str | None: The subject string, or None if inexistent.
        """
        element = self.get_element("//dc:subject")
        if element is None:
            return None
        return element.text

    def set_subject(self, subject: str) -> None:
        """Set the subject of the document.

        (Also available as "self.subject" property.)

        Args:
            subject: The subject string to set.
        """
        element = self.get_element("//dc:subject")
        if element is None:
            element = Element.from_tag("dc:subject")
            self.body.append(element)
        element.text = subject

    @property
    def subject(self) -> str | None:
        """Get or set the subject of a document <dc:subject>.

        Returns: str (or None if inexistent)
        """
        return self.get_subject()

    @subject.setter
    def subject(self, subject: str) -> None:
        return self.set_subject(subject)

    def get_language(self) -> str | None:
        """Get the default language of the document BUT only from the metadata,
        "dc:language".

        Prefer the Document.get_language() function.

        (Also available as "self.language" property.)

        Returns:
            str | None: The language string, or None if inexistent.

        Example::

            >>> document.meta.get_language()
            fr-FR
        """
        element = self.get_element("//dc:language")
        if element is None:
            return None
        return element.text

    def set_language(self, language: str) -> None:
        """Set the default language of the document BUT only in the metadata,
        "dc:language".

        To set globally the default language, prefer the
        Document.set_language() function.

        (Also available as "self.language" property.)

        Args:
            language: The language string to set.

        Example::

            >>> document.meta.set_language('fr-FR')
        """
        language = str(language)
        if not is_RFC3066(language):
            raise TypeError(
                'Language must be "xx" lang or "xx-YY" lang-COUNTRY code (RFC3066)'
            )
        element = self.get_element("//dc:language")
        if element is None:
            element = Element.from_tag("dc:language")
            self.body.append(element)
        element.text = language

    @property
    def language(self) -> str | None:
        """Get or set the default language of the document BUT only in the
        metadata, "dc:language".

        To set globally the default language, prefer the
        Document.language property.

        Returns: str (or None if inexistent)
        """
        return self.get_language()

    @language.setter
    def language(self, language: str) -> None:
        return self.set_language(language)

    def get_creation_date(self) -> datetime | None:
        """Get the creation date of the document.

        (Also available as "self.creation_date" property.)

        Returns:
            datetime | None: The creation date as a datetime object, or None if inexistent.
        """
        element = self.get_element("//meta:creation-date")
        if element is None:
            return None
        creation_date = element.text
        return DateTime.decode(creation_date)

    def set_creation_date(self, date: datetime | None = None) -> None:
        """Set the creation date of the document.

        If provided datetime is None, use current time.

        (Also available as "self.creation_date" property.)

        Args:
            date: The datetime object to set as the creation date.
        """
        element = self.get_element("//meta:creation-date")
        if element is None:
            element = Element.from_tag("meta:creation-date")
            self.body.append(element)
        if date is None:
            date = datetime.now()
        element.text = DateTime.encode(date)

    @property
    def creation_date(self) -> datetime | None:
        """Get or set the date and time when a document was created
        <meta:creation-date>.

        If provided datetime is None, use current time.

        Returns: datetime (or None if inexistent)
        """
        return self.get_creation_date()

    @creation_date.setter
    def creation_date(self, date: datetime | None = None) -> None:
        return self.set_creation_date(date)

    @property
    def print_date(self) -> datetime | None:
        """Get or set the date and time when a document when a document was
        last printed <meta:print-date>

        If provided datetime is None, use current time.

        Returns:
            datetime | None: The print date as a datetime object, or None if inexistent.
        """
        element = self.get_element("//meta:print-date")
        if element is None:
            return None
        date = element.text
        return DateTime.decode(date)

    @print_date.setter
    def print_date(self, date: datetime | None = None) -> None:
        element = self.get_element("//meta:print-date")
        if element is None:
            element = Element.from_tag("meta:print-date")
            self.body.append(element)
        if date is None:
            date = datetime.now()
        element.text = DateTime.encode(date)

    def get_template(self) -> MetaTemplate | None:
        """Get the MetaTemplate "meta:template" element or None.

        Returns:
            A MetaTemplate instance or None.
        """
        element: MetaTemplate | None = self.get_element("//meta:template")  # type: ignore[assignment]
        return element

    @property
    def template(self) -> MetaTemplate | None:
        """Get the MetaTemplate "meta:template" element or None.

        Returns:
            A MetaTemplate instance or None.
        """
        return self.get_template()

    def set_template(
        self,
        date: datetime | None = None,
        href: str = "",
        title: str = "",
    ) -> None:
        """Set or replace the `meta:template` element.

        This specifies the document template used to create the current document.
        Any existing `meta:template` element will be removed.

        Args:
            date: The date and time when the template was used.
            href: The URI for the document template (XLink).
            title: The title of the document template (XLink).
        """
        template = MetaTemplate(date=date, href=href, title=title)
        current = self.template
        if isinstance(current, MetaTemplate):
            current.delete()
        self.body.append(template)

    def get_auto_reload(self) -> MetaAutoReload | None:
        """Get the MetaAutoReload "meta:auto-reload" element or None.

        Returns:
            A MetaAutoReload instance or None.
        """
        element: MetaAutoReload | None = self.get_element("//meta:auto-reload")  # type: ignore[assignment]
        return element

    @property
    def auto_reload(self) -> MetaAutoReload | None:
        """Get the MetaAutoReload "meta:auto-reload" element or None.

        Returns:
            A MetaAutoReload instance or None.
        """
        return self.get_auto_reload()

    def set_auto_reload(self, delay: timedelta, href: str = "") -> None:
        """Set or replace the `meta:auto-reload` element.

        This specifies whether a document is reloaded or replaced by another
        document after a specified period of time. Any existing
        `meta:auto-reload` element will be removed.

        Args:
            delay: The time delay after which the document should auto-reload.
            href: The URL or path to the document to reload or replace with.
        """
        autoreload = MetaAutoReload(delay=delay, href=href)
        current = self.auto_reload
        if isinstance(current, MetaAutoReload):
            current.delete()
        self.body.append(autoreload)

    def get_hyperlink_behaviour(self) -> MetaHyperlinkBehaviour | None:
        """Get the MetaHyperlinkBehaviour "meta:hyperlink-behaviour" element or
        None.

        Returns:
            A MetaHyperlinkBehaviour instance or None.
        """
        element: MetaHyperlinkBehaviour | None = self.get_element(  # type: ignore[assignment]
            "//meta:hyperlink-behaviour"
        )
        return element

    @property
    def hyperlink_behaviour(self) -> MetaHyperlinkBehaviour | None:
        """Get the MetaHyperlinkBehaviour "meta:hyperlink-behaviour" element or
        None.

        Returns:
            A MetaHyperlinkBehaviour instance or None.
        """
        return self.get_hyperlink_behaviour()

    def set_hyperlink_behaviour(
        self,
        target_frame_name: str = "_blank",
        show: str = "replace",
    ) -> None:
        """Set or replace the `meta:hyperlink-behaviour` element.

        This specifies the default behavior for hyperlinks in a document.
        Any existing `meta:hyperlink-behaviour` element will be removed.

        Args:
            target_frame_name: The name of the target frame for the hyperlink.
                Defaults to "_blank" (new window/tab).
            show: Specifies how the target resource is presented.
                Defaults to "replace".
        """
        behaviour = MetaHyperlinkBehaviour(
            target_frame_name=target_frame_name, show=show
        )
        current = self.hyperlink_behaviour
        if isinstance(current, MetaHyperlinkBehaviour):
            current.delete()
        self.body.append(behaviour)

    def get_initial_creator(self) -> str | None:
        """Get the first creator of the document.

        (Also available as "self.initial_creator" property.)

        Returns:
            str | None: The initial creator string, or None if inexistent.

        Example::

            >>> document.meta.get_initial_creator()
            Unknown
        """
        element = self.get_element("//meta:initial-creator")
        if element is None:
            return None
        return element.text

    def set_initial_creator(self, creator: str) -> None:
        """Set the first creator of the document.

        (Also available as "self.initial_creator" property.)

        Args:
            creator: The creator string to set.

        Example::

            >>> document.meta.set_initial_creator("Plato")
        """
        element = self.get_element("//meta:initial-creator")
        if element is None:
            element = Element.from_tag("meta:initial-creator")
            self.body.append(element)
        element.text = creator

    @property
    def initial_creator(self) -> str | None:
        """Get or set the initial creator of a document <meta:initial-creator>.

        Returns: str (or None if inexistent)
        """
        return self.get_initial_creator()

    @initial_creator.setter
    def initial_creator(self, creator: str) -> None:
        return self.set_initial_creator(creator)

    @property
    def printed_by(self) -> str | None:
        """Get or set the name of the last person who printed a document.
        <meta:printed-by>

        Returns:
            str | None: The printed by string, or None if inexistent.
        """
        element = self.get_element("//meta:printed-by")
        if element is None:
            return None
        return element.text

    @printed_by.setter
    def printed_by(self, printed_by: str) -> None:
        element = self.get_element("//meta:printed-by")
        if element is None:
            element = Element.from_tag("meta:printed-by")
            self.body.append(element)
        element.text = printed_by

    def get_keywords(self) -> str | None:
        """Get the keywords of the document. Return the field as-is, without
        any assumption on the keyword separator.

        (Also available as "self.keyword" and "self.keywords" property.)

        Returns:
            str | None: The keywords string, or None if inexistent.
        """
        element = self.get_element("//meta:keyword")
        if element is None:
            return None
        return element.text

    def set_keywords(self, keywords: str) -> None:
        """Set the keywords of the document. Although the name is plural, a str
        string is required, so join your list first.

        (Also available as "self.keyword" and "self.keywords" property.)

        Args:
            keywords: The keywords string to set.
        """
        element = self.get_element("//meta:keyword")
        if element is None:
            element = Element.from_tag("meta:keyword")
            self.body.append(element)
        element.text = keywords

    @property
    def keyword(self) -> str | None:
        """Get or set some keyword(s) keyword pertaining to a document
        <dc:keyword>.

        Returns:
            str | None: The keyword string, or None if inexistent.
        """
        return self.get_keywords()

    @keyword.setter
    def keyword(self, keyword: str) -> None:
        return self.set_keywords(keyword)

    keywords = keyword

    def get_editing_duration(self) -> timedelta | None:
        """Get the time the document was edited, as reported by the generator.

        (Also available as "self.editing_duration" property.)

        Returns:
            timedelta | None: The editing duration as a timedelta object, or None if inexistent.
        """
        element = self.get_element("//meta:editing-duration")
        if element is None:
            return None
        duration = element.text
        return Duration.decode(duration)

    def set_editing_duration(self, duration: timedelta) -> None:
        """Set the time the document was edited.

        (Also available as "self.editing_duration" property.)

        Args:
            duration: The timedelta object representing the editing duration.
        """
        if not isinstance(duration, timedelta):
            raise TypeError("duration must be a timedelta")
        element = self.get_element("//meta:editing-duration")
        if element is None:
            element = Element.from_tag("meta:editing-duration")
            self.body.append(element)
        element.text = Duration.encode(duration)

    @property
    def editing_duration(self) -> timedelta | None:
        """Get or set the total time spent editing a document <meta:editing-
        duration>.

        Returns: timedelta (or None if inexistent)
        """
        return self.get_editing_duration()

    @editing_duration.setter
    def editing_duration(self, duration: timedelta) -> None:
        return self.set_editing_duration(duration)

    def get_editing_cycles(self) -> int | None:
        """Get the number of times the document was edited, as reported by the
        generator.

        (Also available as "self.editing_cycles" property.)

        Returns:
            int | None: The number of editing cycles as an integer, or None if inexistent.
        """
        element = self.get_element("//meta:editing-cycles")
        if element is None:
            return None
        cycles = element.text
        return int(cycles)

    def set_editing_cycles(self, cycles: int) -> None:
        """Set the number of times the document was edited.

        (Also available as "self.editing_cycles" property.)

        Args:
            cycles: The number of editing cycles to set.
        """
        if not isinstance(cycles, int):
            raise TypeError("cycles must be an int")
        if cycles < 1:
            raise ValueError("cycles must be a positive int")
        element = self.get_element("//meta:editing-cycles")
        if element is None:
            element = Element.from_tag("meta:editing-cycles")
            self.body.append(element)
        element.text = str(cycles)

    @property
    def editing_cycles(self) -> int | None:
        """Get or set the number of times a document has been edited
        <meta:editing-cycles>.

        When a document is created, this value is set to 1. Each time
        a document is saved, the editing-cycles number is incremented by 1.

        Returns: int (or None if inexistent)
        """
        return self.get_editing_cycles()

    @editing_cycles.setter
    def editing_cycles(self, cycles: int) -> None:
        return self.set_editing_cycles(cycles)

    @property
    def generator(self) -> str | None:
        """Get or set the signature of the software that generated this
        document.

        Returns:
            str | None: The generator string, or None if inexistent.

        Example::

            >>> document.meta.generator
            KOffice/2.0.0
            >>> document.meta.generator = "Odfdo experiment"
        """
        element = self.get_element("//meta:generator")
        if element is None:
            return None
        return element.text

    @generator.setter
    def generator(self, generator: str) -> None:
        element = self.get_element("//meta:generator")
        if element is None:
            element = Element.from_tag("meta:generator")
            self.body.append(element)
        element.text = generator
        self._generator_modified = True

    def get_generator(self) -> str | None:
        """Get the signature of the software that generated this document.

        (Also available as "self.generator" property.)

        Returns:
            str (or None if inexistent).

        Example::

            >>> document.meta.get_generator()
            KOffice/2.0.0
        """
        return self.generator

    def set_generator(self, generator: str) -> None:
        """Set the signature of the software that generated this document.

        (Also available as "self.generator" property.)

        Args:
            generator: The generator string to set.

        Example::

            >>> document.meta.set_generator("Odfdo experiment")
        """
        self.generator = generator

    def set_generator_default(self) -> None:
        """Set the signature of the software that generated this document to
        ourself.

        Example::

            >>> document.meta.set_generator_default()
        """
        if not self._generator_modified:
            self.generator = GENERATOR

    def get_statistic(self) -> dict[str, int] | None:
        """Get the statistics about a document, "meta:document-statistic" tag.

        (Also available as "self.statistic" property.)

        Returns:
            A dict (or None if inexistent).

        Example::

            >>> document.get_statistic():
            {'meta:table-count': 1,
             'meta:image-count': 2,
             'meta:object-count': 3,
             'meta:page-count': 4,
             'meta:paragraph-count': 5,
             'meta:word-count': 6,
             'meta:character-count': 7,
             'meta:non-whitespace-character-count': 3}
        """
        element = self.get_element("//meta:document-statistic")
        if element is None:
            return None
        statistic = {}
        for key, value in element.attributes.items():
            statistic[to_str(key)] = int(value)
        return statistic

    def set_statistic(self, statistic: dict[str, int]) -> None:
        """Set the statistics about a document, "meta:document-statistic" tag.

        (Also available as "self.statistic" property.)

        Args:
            statistic: The statistics dictionary to set.

        Example::

            >>> statistic = {'meta:table-count': 1,
                             'meta:image-count': 2,
                             'meta:object-count': 3,
                             'meta:page-count': 4,
                             'meta:paragraph-count': 5,
                             'meta:word-count': 6,
                             'meta:character-count': 7,
                             'meta:non-whitespace-character-count': 3}
            >>> document.meta.set_statistic(statistic)
        """
        if not isinstance(statistic, dict):
            raise TypeError("Statistic must be a dict")
        element = self.get_element("//meta:document-statistic")
        if element is None:
            element = Element.from_tag("meta:document-statistic")
            self.body.append(element)
        for key, value in statistic.items():
            try:
                ivalue = int(value)
            except ValueError as e:
                msg = f"Statistic value must be a int: {key}:{value!r}"
                raise TypeError(msg) from e
            element.set_attribute(to_str(key), str(ivalue))

    @property
    def statistic(self) -> dict[str, int] | None:
        """Get or set the statistics about a document, "meta:document-
        statistic" tag.

        Returns:
            dict[str, int] | None: The statistics dictionary, or None if inexistent.

        Example::

            >>> document.get_statistic():
            {'meta:table-count': 1,
             'meta:image-count': 2,
             'meta:object-count': 3,
             'meta:page-count': 4,
             'meta:paragraph-count': 5,
             'meta:word-count': 6,
             'meta:character-count': 7,
             'meta:non-whitespace-character-count':3}
        """
        return self.get_statistic()

    @statistic.setter
    def statistic(self, statistic: dict[str, int]) -> None:
        return self.set_statistic(statistic)

    def get_user_defined_metadata(
        self,
    ) -> dict[str, Decimal | datetime | dtdate | timedelta | bool | str]:
        """Get all additional user-defined metadata for a document.

        (Also available as "self.user_defined_metadata" property.)

        Returns:
           A dict of str/value mapping.

        Value types can be: Decimal, datetime, date, timedelta, bool or str.
        """
        return {
            data.name: data.value
            for data in cast(
                list[MetaUserDefined], self.get_elements("//meta:user-defined")
            )
        }

    def _user_defined_metadata_list(
        self,
    ) -> list[dict[str, Decimal | datetime | dtdate | timedelta | bool | str]]:
        """Internal helper to retrieve all user-defined metadata as a sorted list of dictionaries.

        Returns:
            list[dict[str, Any]]: A sorted list of dictionaries, each
                representing a user-defined metadata field.
        """
        user_defined = [
            data.as_dict()
            for data in cast(
                list[MetaUserDefined], self.get_elements("//meta:user-defined")
            )
        ]
        return sorted(user_defined, key=itemgetter("meta:name"))

    def clear_user_defined_metadata(self) -> None:
        """Remove all user-defined metadata fields from the document.

        Iteratively deletes all `meta:user-defined` elements.
        """
        while True:
            element = self.get_element("//meta:user-defined")
            if isinstance(element, MetaUserDefined):
                element.delete()
                continue
            break

    @property
    def user_defined_metadata(
        self,
    ) -> dict[str, Decimal | datetime | dtdate | timedelta | bool | str]:
        """Get or set all additional user-defined metadata for a document.

        Return a dict of str/value mapping.

        Value types can be: Decimal, datetime, date, timedelta, bool or str.
        """
        return self.get_user_defined_metadata()

    @user_defined_metadata.setter
    def user_defined_metadata(self, metadata: dict[str, Any]) -> None:
        self.clear_user_defined_metadata()
        for key, val in metadata.items():
            self.set_user_defined_metadata(name=key, value=val)

    def _user_defined_metadata_by_name(self, name: str) -> MetaUserDefined | None:
        """Internal helper to find a specific user-defined metadata element by name.

        Args:
            name: The name of the user-defined metadata field to find.

        Returns:
            MetaUserDefined | None: The `MetaUserDefined` element if found,
                otherwise `None`.
        """
        for item in cast(
            list[MetaUserDefined], self.get_elements("//meta:user-defined")
        ):
            if item.name == name:
                return item
        return None

    def get_user_defined_metadata_of_name(self, keyname: str) -> dict[str, Any] | None:
        """Return the content of the user defined metadata of that name. Return
        None if no name matches or a dic of fields.

        Args:
            keyname: The name of the user-defined metadata field to find.

        Returns:
            dict[str, Any]: A dict with keys "name", "value", "value_type", "text".
        """
        item = self._user_defined_metadata_by_name(keyname)
        if item is None:
            return None
        return item.as_dict_full()

    def set_user_defined_metadata(
        self,
        name: str,
        value: bool
        | int
        | float
        | Decimal
        | datetime
        | dtdate
        | str
        | timedelta
        | None,
    ) -> None:
        """Set a user defined metadata of that name and value.

        If value is None, any existing metadata of that name is deleted.

        Args:
            name: The name of the user-defined metadata field.
            value: The value to set for the metadata field.
                Can be a boolean, int, float, Decimal, datetime, date, string,
                timedelta, or None for deletion.
        """
        if value is None:
            self.delete_user_defined_metadata_of_name(name)
            return
        value_type = MetaUserDefined._value_to_value_type(value)
        # Already the same element ?
        metadata = self._user_defined_metadata_by_name(name)
        if metadata is None:
            metadata = MetaUserDefined(name=name, value_type=value_type, value=value)
            self.body.append(metadata)
        else:
            metadata.value_type = value_type
            metadata.value = value

    def delete_user_defined_metadata_of_name(self, name: str) -> None:
        """Delete all user-defined metadata fields with a specific name.

        Args:
            name: The name of the user-defined metadata field to delete.
        """
        while True:
            metadata = self._user_defined_metadata_by_name(name)
            if not metadata:
                return
            metadata.delete()

    def as_dict(self, full: bool = False) -> dict[str, Any]:
        """Return the metadata of the document as a Python dict.

        if 'full' is True, export also the keys with no value assigned.

        Args:
            full: If True, exports also the keys with no value assigned.

        Returns:
            The metadata of the document as a Python dict.
        """

        def _stats() -> dict[str, int]:
            doc_stats = self.statistic
            if doc_stats is None:
                msg = "Document statistics not found"
                raise LookupError(msg)
            return {
                key: doc_stats.get(key, 0)
                for key in (
                    "meta:table-count",
                    "meta:image-count",
                    "meta:object-count",
                    "meta:page-count",
                    "meta:paragraph-count",
                    "meta:word-count",
                    "meta:character-count",
                    "meta:non-whitespace-character-count",
                )
            }

        def _meta_template() -> dict[str, Any] | None:
            template = self.template
            if template is None:
                return None
            return template.as_dict()

        def _meta_reload() -> dict[str, Any] | None:
            reload = self.auto_reload
            if reload is None:
                return None
            return reload.as_dict()

        def _meta_behaviour() -> dict[str, Any] | None:
            behaviour = self.hyperlink_behaviour
            if behaviour is None:
                return None
            return behaviour.as_dict()

        meta_data: dict[str, Any] = {
            "meta:creation-date": self.creation_date,
            "dc:date": self.date,
            "meta:editing-duration": self.editing_duration,
            "meta:editing-cycles": self.editing_cycles,
            "meta:document-statistic": _stats(),
            "meta:generator": self.generator,
            "dc:title": self.title,
            "dc:description": self.description,
            "dc:creator": self.creator,
            "meta:keyword": self.keyword,
            "dc:subject": self.subject,
            "dc:language": self.language,
            "meta:initial-creator": self.initial_creator,
            "meta:print-date": self.print_date,
            "meta:printed-by": self.printed_by,
            "meta:template": _meta_template(),
            "meta:auto-reload": _meta_reload(),
            "meta:hyperlink-behaviour": _meta_behaviour(),
            "meta:user-defined": self._user_defined_metadata_list(),
        }

        if not full:
            meta_data = {key: val for key, val in meta_data.items() if val}
        return meta_data

    def _as_json_dict(self, full: bool = False) -> dict[str, Any]:
        """Internal method to prepare metadata as a JSON-compatible dictionary.

        This converts various Python types (datetime, timedelta, Decimal) into
        their string representations suitable for JSON serialization.

        Args:
            full (bool): If True, exports all keys including those with no value assigned.

        Returns:
            dict[str, Any]: A dictionary of metadata ready for JSON serialization.
        """

        def _convert(data: dict[str, Any]) -> None:
            for key, val in data.items():
                if isinstance(val, datetime):
                    data[key] = DateTime.encode(val)
                elif isinstance(val, timedelta):
                    data[key] = Duration.encode(val)
                elif isinstance(val, Decimal):
                    data[key] = json.loads(str(Decimal(val)))

        meta_data: dict[str, Any] = self.as_dict(full=full)
        user_defined = meta_data.get("meta:user-defined", [])
        for item in user_defined:
            _convert(item)
        meta_data["meta:user-defined"] = user_defined
        meta_template = meta_data.get("meta:template")
        if meta_template is not None:
            _convert(meta_template)
            meta_data["meta:template"] = meta_template
        reload = meta_data.get("meta:auto-reload")
        if reload is not None:
            _convert(reload)
            meta_data["meta:auto-reload"] = reload
        meta_data["meta:user-defined"] = user_defined

        _convert(meta_data)
        return meta_data

    def as_json(self, full: bool = False) -> str:
        """Return the metadata of the document as a JSON string.

        if 'full' is True, export also the keys with no value assigned.

        Args:
            full:  boolean

        Returns:
            The metadata of the document as a JSON string.
        """
        return json.dumps(
            self._as_json_dict(full=full),
            ensure_ascii=False,
            sort_keys=False,
            indent=4,
        )

    def as_text(self, no_user_defined_msg: str = "") -> str:
        """Return meta information as text, with some formatting for
        printing.

        Args:
            no_user_defined_msg: customize the string when no user defined
                data is present.

        Returns:
            The metadata of the document as a text.
        """
        data = self._as_json_dict(full=False)
        result: list[str] = []

        def _append_info(name: str, key: str) -> None:
            value = data.get(key)
            if value:
                result.append(f"{name}: {value}")

        _append_info("Title", "dc:title")
        _append_info("Subject", "dc:subject")
        _append_info("Description", "dc:description")
        _append_info("Language", "dc:language")
        _append_info("Modification date", "dc:date")
        _append_info("Creation date", "meta:creation-date")
        _append_info("Creator", "dc:creator")
        _append_info("Initial creator", "meta:initial-creator")
        _append_info("Keyword", "meta:keyword")
        _append_info("Editing duration", "meta:editing-duration")
        _append_info("Editing cycles", "meta:editing-cycles")
        _append_info("Generator", "meta:generator")

        result.append("Statistic:")
        statistic = data.get("meta:document-statistic", {})
        for name, value in statistic.items():
            result.append(f"  - {name[5:].replace('-', ' ').capitalize()}: {value}")

        user_metadata = data.get("meta:user-defined", [])
        if user_metadata:
            msg = ""
        else:
            msg = f" {no_user_defined_msg}"
        result.append(f"User defined metadata:{msg}")
        for item in user_metadata:
            result.append(f"  - {item['meta:name']}: {item['value']}")

        return "\n".join(result)

    @staticmethod
    def _complete_stats(
        current_stats: dict[str, int],
        imported_stats: dict[str, int] | None,
    ) -> dict[str, int]:
        """Internal helper to merge and complete document statistics.

        It takes current statistics and imported statistics, filling in missing
        values with 0 and prioritizing imported values.

        Args:
            current_stats (dict[str, int]): The current document statistics.
            imported_stats (dict[str, int] | None): The statistics to import,
                or None.

        Returns:
            dict[str, int]: A merged dictionary of statistics.
        """
        if imported_stats is None:
            imported_stats = {}
            current_stats = {}
        new: dict[str, int] = {}
        for key in (
            "meta:table-count",
            "meta:image-count",
            "meta:object-count",
            "meta:page-count",
            "meta:paragraph-count",
            "meta:word-count",
            "meta:character-count",
            "meta:non-whitespace-character-count",
        ):
            new[key] = imported_stats.get(key, current_stats.get(key, 0))
        return new

    def from_dict(self, data: dict[str, Any]) -> None:
        """Set the metadata of the document from a Python dict.

        The loaded metadata are merged with the existing metadata.
        If the new value of a key is None:
            - meta:creation-date: use current time,
            - dc:date: use creation date,
            - meta:editing-duration: set to zero,
            - meta:editing-cycles: set to 1,
            - meta:generator: use odfdo generator string.
            Other keys (not mandatory keys): remove key/value pair from
            metadata.

        Args:
            data: A dictionary of metadata.
        """

        def _value_delete(key: str) -> Any:
            value = data.get(key, current.get(key))
            if value is None:
                child = self.get_element(f"//{key}")
                if child is not None:
                    self.delete_element(child)
            return value

        current = self.as_dict()
        new_stats = self._complete_stats(
            current["meta:document-statistic"],
            data.get("meta:document-statistic", {}),
        )
        # mandatory
        self.statistic = new_stats

        key = "meta:creation-date"
        creation_date = data.get(key, current.get(key))
        if creation_date is None:
            creation_date = datetime.now().replace(microsecond=0)
        self.creation_date = creation_date

        key = "dc:date"
        dc_date = data.get(key, current.get(key))
        if dc_date is None:
            dc_date = creation_date
        if dc_date < creation_date:
            dc_date = creation_date
        max_editing = dc_date - creation_date
        self.date = dc_date

        key = "meta:editing-duration"
        editing_duration = data.get(key, current.get(key))
        if editing_duration is None:
            editing_duration = timedelta(0)
        if editing_duration > max_editing:
            editing_duration = max_editing
        self.editing_duration = editing_duration

        key = "meta:editing-cycles"
        editing_cycles = data.get(key, current.get(key))
        if editing_cycles is None:
            editing_cycles = 1
        self.editing_cycles = max(editing_cycles, 1)

        key = "meta:generator"
        generator = data.get(key, current.get(key))
        if not generator:
            generator = GENERATOR
        self.generator = generator

        # not mandatory values
        # - if not in imported: keep original value
        # - if imported is None: erase original value

        key = "dc:title"
        value = _value_delete(key)
        if value is not None:
            self.title = value

        key = "dc:description"
        value = _value_delete(key)
        if value is not None:
            self.description = value

        key = "dc:creator"
        value = _value_delete(key)
        if value is not None:
            self.creator = value

        key = "meta:keyword"
        value = _value_delete(key)
        if value is not None:
            self.keyword = value

        key = "dc:subject"
        value = _value_delete(key)
        if value is not None:
            self.subject = value

        key = "dc:language"
        value = _value_delete(key)
        if value is not None:
            self.language = value

        key = "meta:initial-creator"
        value = _value_delete(key)
        if value is not None:
            self.initial_creator = value

        key = "meta:print-date"
        value = _value_delete(key)
        if value is not None:
            self.print_date = value

        key = "meta:printed-by"
        value = _value_delete(key)
        if value is not None:
            self.printed_by = value

        key = "meta:template"
        value = _value_delete(key)
        if value is not None:
            self.set_template(
                date=value["meta:date"],
                href=value["xlink:href"],
                title=value["xlink:title"],
            )

        key = "meta:auto-reload"
        value = _value_delete(key)
        if value is not None:
            self.set_auto_reload(
                delay=value["meta:delay"],
                href=value["xlink:href"],
            )

        key = "meta:hyperlink-behaviour"
        value = _value_delete(key)
        if value is not None:
            self.set_hyperlink_behaviour(
                target_frame_name=value["office:target-frame-name"],
                show=value["xlink:show"],
            )

        key = "meta:user-defined"
        if key in data:
            value = data[key]
            if value is None:
                self.clear_user_defined_metadata()
            else:
                data_list = self._user_defined_metadata_list()
                current_dict = {d["meta:name"]: d for d in data_list}
                current_value = {d["meta:name"]: d for d in value}
                current_dict.update(current_value)
                new_ud = {
                    v["meta:name"]: v["value"]
                    for v in current_dict.values()
                    if v["value"] is not None
                }
                self.user_defined_metadata = new_ud  # type: ignore[assignment]

    def strip(
        self,
        generator: str = GENERATOR,
        creation_date: datetime | None = None,
    ) -> None:
        """Strip metadata to their minimal content.

        By default the new metadata values are:
            - meta:creation-date: use current time,
            - dc:date: use creation date,
            - meta:editing-duration: set to zero,
            - meta:editing-cycles: set to 1,
            - meta:generator: use odfdo generator string.
            - all meta:document-statistic values to 0.

        All user defined metadata are removed.

        Args:
            generator: String for the meta:generator field.
            creation_date: Datetime or None, meta:creation-date value.
        """

        self.body.clear()
        self.statistic = self._complete_stats({}, None)
        if creation_date is None:
            self.creation_date = datetime.now().replace(microsecond=0)
        else:
            self.creation_date = creation_date.replace(microsecond=0)
        self.date = self.creation_date
        self.editing_duration = timedelta(0)
        self.editing_cycles = 1
        self.generator = generator
        self.clear_user_defined_metadata()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Classes for ODF document variables declarations.

This module provides classes for managing variable declaration
"text:variable-decl" and variable declaration container
"text:variable-decls"."""

from __future__ import annotations

from typing import Any, Union, cast

from .element import FIRST_CHILD, Element, PropDef, register_element_class


class VarDeclMixin(Element):
    """Mixin class for elements that can contain variable declarations.

    This mixin provides methods to access and manipulate "text:variable-decls"
    and "text:variable-decl".

    Used by the following classes:
        - "office:chart"
        - "office:drawing"
        - "office:presentation"
        - "office:spreadsheet"
        - "office:text"
        - "style:footer"
        - "style:footer-first"
        - "style:footer-left"
        - "style:header"
        - "style:header-first"
        - "style:header-left"
    """

    def get_variable_decls(self) -> VarDecls:
        """Returns the container for variable declarations.

        If the container is not found, it is created within the document body.

        Returns:
            VarDecls: The VarDecls instance (container for variable declarations).

        Raises:
            ValueError: If the document body is empty and a new container cannot be inserted.
        """
        variable_decls = self.get_element("//text:variable-decls")
        if variable_decls is None:
            body = self.document_body
            if not body:
                raise ValueError("Empty document.body")
            body.insert(Element.from_tag("text:variable-decls"), FIRST_CHILD)
            variable_decls = body.get_element("//text:variable-decls")

        return cast(VarDecls, variable_decls)

    def get_variable_decl_list(self) -> list[VarDecls]:
        """Returns all variable declarations as a list.

        Returns:
            list[VarDecls]: A list of all VarDecls instances that are descendants of this element.
        """
        return cast(
            list[VarDecls], self._filtered_elements("descendant::text:variable-decl")
        )

    def get_variable_decl(self, name: str, position: int = 0) -> VarDecls | None:
        """Returns a single variable declaration that matches the specified criteria.

        Args:
            name: The name of the variable declaration to retrieve.
            position: The 0-based index of the matching variable declaration to return.

        Returns:
            VarDecls | None: A VarDecls instance, or None if no declaration matches the criteria.
        """
        return cast(
            Union[None, VarDecls],
            self._filtered_element(
                "descendant::text:variable-decl", position, text_name=name
            ),
        )


class VarDecls(Element):
    """A container for variable declarations, "text:variable-decls".

    This element groups all the 'text:variable-decl' elements in the
    document.
    """

    _tag = "text:variable-decls"


class VarDecl(Element):
    """A variable declaration, "text:variable-decl".

    This element defines a variable by its name and value type.

    Attributes:
        name (str): The unique name of the variable.
        value_type (str): The ODF value type (e.g., 'string', 'float').
    """

    _tag = "text:variable-decl"
    _properties = (
        PropDef("name", "text:name"),
        PropDef("value_type", "office:value-type"),
    )

    def __init__(
        self,
        name: str | None = None,
        value_type: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the VarDecl element.

        Args:
            name: The name of the variable.
            value_type: The ODF value type.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name:
                self.name = name
            if value_type:
                self.value_type = value_type


VarDecl._define_attribut_property()


register_element_class(VarDecls)
register_element_class(VarDecl)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.user_field import UserFieldGet, UserFieldInput

ZOE = "你好 Zoé"
CHAMP = "Champêtre"


@pytest.fixture
def document(samples) -> Iterable[Document]:
    document = Document(samples("variable.odt"))
    yield document


def test_user_field_get_class():
    field = UserFieldGet()
    assert isinstance(field, UserFieldGet)


def test_user_field_get_from_tag():
    field = Element.from_tag(
        '<text:user-field-get text:name="some_name" '
        'style:data-style-name="some_style">'
        "</text:user-field-get>"
    )
    assert isinstance(field, UserFieldGet)


def test_user_field_get_style():
    field = UserFieldGet(name="some_name", style="some_style")
    assert field.style == "some_style"


def test_user_field_input_class():
    field = UserFieldInput()
    assert isinstance(field, UserFieldInput)


def test_create_user_field_get():
    user_field_get = UserFieldGet(ZOE, value=42)
    expected = (
        f'<text:user-field-get text:name="{ZOE}" '
        'office:value-type="float" calcext:value-type="float" '
        'office:value="42" calcext:value="42">'
        "42"
        "</text:user-field-get>"
    )
    assert user_field_get.serialize() == expected


def test_create_user_field_input():
    user_field_input = UserFieldInput(ZOE, value=42)
    expected = (
        f'<text:user-field-input text:name="{ZOE}" '
        'office:value-type="float" calcext:value-type="float" '
        'office:value="42" calcext:value="42">'
        "42"
        "</text:user-field-input>"
    )
    assert user_field_input.serialize() == expected


def test_get_user_field_get(document):
    body = document.body
    value = body.get_user_field_value(CHAMP)
    assert value


def test_get_user_field_get_none(document):
    body = document.body
    value = body.get_user_field_value("unknown")
    assert value is None



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>

import pytest

from odfdo.style import make_table_cell_border_string


def test_default_border():
    expected = "0.06pt solid #000000"
    assert make_table_cell_border_string() == expected


def test_border_bold():
    expected = "0.30pt solid #000000"
    assert make_table_cell_border_string(thick=0.3) == expected


def test_border_bold2():
    expected = "0.30pt solid #000000"
    assert make_table_cell_border_string(thick="0.30pt") == expected


def test_border_int():
    expected = "0.06pt solid #000000"
    assert make_table_cell_border_string(thick=6) == expected


def test_border_custom():
    expected = "0.1cm dotted #FF0000"
    result = make_table_cell_border_string(
        thick="0.1cm ",
        line=" dotted",
        color="#FF0000 ",
    )
    assert result == expected


def test_border_color_raise():
    arg = (None, None, "bug")
    with pytest.raises(ValueError):
        make_table_cell_border_string(arg)


def test_border_color_name():
    expected = "0.06pt solid #D3D3D3"
    assert make_table_cell_border_string(color="lightgrey") == expected


def test_border_color_tuple():
    expected = "0.06pt solid #808080"
    assert make_table_cell_border_string(color=(128, 128, 128)) == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Romain Gauthier <romain@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
"""Span class for "text:span" tag."""

from __future__ import annotations

from typing import Any

from .element import (
    Element,
    PropDef,
    register_element_class,
)
from .mixin_link import LinkMixin
from .mixin_md import MDParagraph, MDSpan
from .mixin_paragraph import ParaMixin
from .mixin_paragraph_formatted import ParaFormattedTextMixin
from .note import NoteMixin
from .user_field import UserDefinedMixin


class Span(
    MDSpan,
    MDParagraph,
    UserDefinedMixin,
    LinkMixin,
    ParaFormattedTextMixin,
    ParaMixin,
    NoteMixin,
    Element,
):
    """A span tag (styled text in paragraph), "text:span".

    This element is used for inline text styling within a paragraph. It can
    contain text content and is often associated with a text style.

    Inherits from various mixins to provide markdown, paragraph, and note
    functionalities.
    """

    _tag = "text:span"
    _properties = (
        PropDef("style", "text:style-name"),
        PropDef("class_names", "text:class-names"),
    )

    def __init__(
        self,
        text: str | None = None,
        style: str | None = None,
        formatted: bool = True,
        **kwargs: Any,
    ) -> None:
        """Create a span element "text:span" of the given style containing the
        optional given text.

        If "formatted" is True (the default), the given text is appended with <CR>,
        <TAB> and multiple spaces replaced by ODF corresponding tags.

        Args:
            text: The text content for the span.
            style: The style name for the span.
            formatted: If True, special characters in `text` are
                replaced by ODF corresponding tags. Defaults to True.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if text:
                if formatted:
                    self.text = ""
                    self.append_plain_text(text)
                else:
                    self.text = self._unformatted(text)
            if style:
                self.style = style

    def __str__(self) -> str:
        return self.inner_text


Span._define_attribut_property()

register_element_class(Span)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>

from odfdo.column import Column


def test_column_init():
    column = Column()
    expected = "<table:table-column/>"
    assert column.serialize() == expected


def test_column_repr():
    column = Column()
    expected = "<Column x=None>"
    assert repr(column) == expected


def test_column_default_cell_style():
    column = Column(default_cell_style="A Style")
    expected = '<table:table-column table:default-cell-style-name="A Style"/>'
    assert column.serialize() == expected


def test_column_style():
    column = Column(style="A Style")
    expected = '<table:table-column table:style-name="A Style"/>'
    assert column.serialize() == expected


def test_column_all():
    column = Column(style="co1", default_cell_style="Standard", repeated=3)
    expected = (
        "<table:table-column "
        'table:default-cell-style-name="Standard" '
        'table:number-columns-repeated="3" '
        'table:style-name="co1"/>'
    )
    assert column.serialize() == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Luis Belmar-Letelier <luis@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>

from odfdo.document import Document
from odfdo.element import Element


def test_insert_style_paragraph(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        '<style:style style:name="custom" '
        'style:display-name="custom" '
        'style:family="paragraph" '
        'style:parent-style-name="Text">'
        '<style:paragraph-properties fo:margin-left="2cm"/>'
        '<style:text-properties fo:color="#808080" loext:opacity="100%" '
        'fo:font-size="16pt" fo:font-style="normal" '
        'style:text-underline-style="solid" '
        'style:text-underline-width="auto" '
        'style:text-underline-color="font-color" '
        'fo:font-weight="bold"/>'
        "</style:style>"
    )
    nb_styles = len(document.get_styles("paragraph"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("paragraph"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("paragraph"))
    assert result == expected


def test_insert_style_date(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        "<number:date-style "
        'style:name="nr-nl-date" '
        'number:automatic-order="true">'
        '<number:day number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:month number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:year number:style="long"/>'
        "</number:date-style>"
    )
    nb_styles = len(document.get_styles("date"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("date"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("date"))
    assert result == expected
    result2 = document.get_style("date", "nr-nl-date")
    assert result2


def test_insert_style_currency(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency-gt0" '
        'style:volatile="true">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2"'
        ' number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "</number:currency-style>"
    )
    nb_styles = len(document.get_styles("currency"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("currency"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("currency"))
    assert result == expected
    result2 = document.get_style("currency", "nr-nl-currency-gt0")
    assert result2


def test_insert_style_currency_2(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2" '
        'number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "<number:text>-</number:text>"
        "<style:map "
        'style:condition="value()&gt;=0" '
        'style:apply-style-name="nr-nl-currency-gt0"/>'
        "</number:currency-style>"
    )
    nb_styles = len(document.get_styles("currency"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("currency"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("currency"))
    assert result == expected
    result2 = document.get_style("currency", "nr-nl-currency")
    assert result2


def test_insert_style_composite_1(samples):
    document = Document(samples("example.odt"))
    style_date = Element.from_tag(
        "<number:date-style "
        'style:name="nr-nl-date" '
        'number:automatic-order="true">'
        '<number:day number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:month number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:year number:style="long"/>'
        "</number:date-style>"
    )
    document.insert_style(style_date)
    style_cur = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2" '
        'number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "<number:text>-</number:text>"
        "<style:map "
        'style:condition="value()&gt;=0" '
        'style:apply-style-name="nr-nl-currency-gt0"/>'
        "</number:currency-style>"
    )
    document.insert_style(style_cur)
    style = Element.from_tag(
        "<style:style "
        'style:name="nldate" '
        'style:family="table-cell" '
        'style:parent-style-name="Default" '
        'style:data-style-name="nr-nl-date"/>'
    )
    nb_styles = len(document.get_styles("table-cell"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("table-cell"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("table-cell"))
    assert result == expected


def test_insert_style_composite_2(samples):
    document = Document(samples("example.odt"))
    style_date = Element.from_tag(
        "<number:date-style "
        'style:name="nr-nl-date" '
        'number:automatic-order="true">'
        '<number:day number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:month number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:year number:style="long"/>'
        "</number:date-style>"
    )
    document.insert_style(style_date)
    style_cur = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2" '
        'number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "<number:text>-</number:text>"
        "<style:map "
        'style:condition="value()&gt;=0" '
        'style:apply-style-name="nr-nl-currency-gt0"/>'
        "</number:currency-style>"
    )
    document.insert_style(style_cur)
    style = Element.from_tag(
        "<style:style "
        'style:name="nlcurrency" '
        'style:family="table-cell" '
        'style:parent-style-name="Default" '
        'style:data-style-name="nr-nl-currency"/>'
    )
    nb_styles = len(document.get_styles("table-cell"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("table-cell"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("table-cell"))
    assert result == expected



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
from __future__ import annotations

import re
from collections import namedtuple
from collections.abc import Iterable

import pytest

from odfdo.const import ODF_CONTENT
from odfdo.container import Container
from odfdo.element import (
    FIRST_CHILD,
    NEXT_SIBLING,
    PREV_SIBLING,
    Element,
    _decode_qname,
    _uri_to_prefix,
    _xpath_text_descendant_no_annotation,
    register_element_class,
    xpath_compile,
)
from odfdo.named_range import NamedRange
from odfdo.paragraph import Paragraph
from odfdo.section import Section
from odfdo.xmlpart import XmlPart

SPECIAL_CHARS = 'using < & " characters'

Sample = namedtuple("Sample", ["container", "content", "para", "anno", "span"])


class DummyElement(Element):
    _tag = "office:dummy1"


@pytest.fixture
def sample(samples) -> Iterable[Sample]:
    container = Container(samples("example.odt"))
    content = XmlPart(ODF_CONTENT, container)
    para = content.get_element("//text:p[1]")
    anno = content.get_element("//office:annotation[1]")
    span = ""
    yield Sample(container=container, content=content, para=para, anno=anno, span=span)


@pytest.fixture
def span_styles(samples) -> Iterable[Sample]:
    container = Container(samples("span_style.odt"))
    content = XmlPart(ODF_CONTENT, container)
    para = content.get_element("//text:p")
    anno = ""
    span = para.get_element("//text:span")
    yield Sample(container=container, content=content, para=para, anno=anno, span=span)


def test_element_init_completude_vs_registry():
    import odfdo

    all_imported = set(odfdo.__all__)
    registry = odfdo.element._class_registry
    for klass in registry.values():
        assert klass.__name__ in all_imported


def test_create_simple():
    data = "<p>Template Element</p>"
    element = Element.from_tag(data)
    assert element.serialize() == data


def test_create_namespace():
    data = "<text:p>Template Element</text:p>"
    element = Element.from_tag(data)
    assert element.serialize() == data


def test_create_qname():
    element = Element.from_tag("text:p")
    assert element.serialize() == "<text:p/>"


def test_decode_qname_bad():
    with pytest.raises(ValueError):
        _decode_qname("hip:hop")


def test_uri_to_prefix_bad():
    with pytest.raises(ValueError):
        _uri_to_prefix("wrong_uri")


def test_etext_parent_1():
    element = Element.from_tag('<text:p style="aa"/>')
    lst = element.xpath("//@style")
    etext = lst[0]
    assert isinstance(etext.parent, Paragraph)


def test_etext_parent_2():
    element = Element.from_tag('<text:p style="aa"/>')
    lst = element.xpath("//@style")
    etext = lst[0]
    etext._EText__parent = None
    assert etext.parent is None


def test_bad_python_element():
    with pytest.raises(TypeError):
        Element("<text:p/>")


def test_bad_element_from_tag():
    with pytest.raises(TypeError):
        Element(tag_or_elem=[])


def test_make_etree_element_bad_1():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(TypeError):
        element._make_etree_element([])


def test_make_etree_element_bad_2():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(ValueError):
        element._make_etree_element(" ")


def test_generic_attrib_getter_bad_1():
    element = Element.from_tag("<text:p/>")
    fct = element._generic_attrib_getter("atr", "fam")
    assert fct(element) is None


def test_generic_attrib_getter_bad_2():
    element = Element.from_tag("<text:p/>")
    element.family = "nope"
    fct = element._generic_attrib_getter("atr", "fam")
    assert fct(element) is None


def test_generic_attrib_setter_bad_1():
    element = Element.from_tag("<text:p/>")
    fct = element._generic_attrib_setter("atr", "fam")
    assert fct(element, "value") is None


def test_generic_attrib_setter_bad_2():
    element = Element.from_tag("<text:p/>")
    element.family = "nope"
    fct = element._generic_attrib_setter("atr", "fam")
    assert fct(element, "value") is None


def test_make_before_regex_bad():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(ValueError):
        element._make_before_regex(None, None)


def test_search_negative_position_bad_1():
    element = Element.from_tag("<text:p/>")
    pattern = re.compile("abc")
    with pytest.raises(ValueError):
        element._search_negative_position([], pattern)


def test_search_negative_position_bad_2():
    element = Element.from_tag("<text:p>abc</text:p>")
    pattern = re.compile("abc")
    with pytest.raises(TypeError):
        element._search_negative_position([element], pattern)


def test_search_positive_position_bad_1():
    element = Element.from_tag("<text:p/>")
    pattern = re.compile("abc")
    with pytest.raises(ValueError):
        element._search_positive_position([], pattern, 0)


def test_search_positive_position_bad_2():
    element = Element.from_tag("<text:p>abc</text:p>")
    pattern = re.compile("abc")
    with pytest.raises(TypeError):
        element._search_positive_position(["abc", element], pattern, 1)


def test__insert_find_text_bad():
    current = Element.from_tag("<text:p>abcde</text:p>")
    element = Element.from_tag("<text:p>fghij</text:p>")
    with pytest.raises(ValueError):
        element._insert_find_text(
            current._Element__element,
            element._Element__element,
            "x",
            "y",
            200,
            _xpath_text_descendant_no_annotation,
        )


def test_elements_repeated_sequence_1():
    para = Paragraph("aaa")
    x = xpath_compile(".")
    result = para.elements_repeated_sequence(x, "b")
    assert result == [(0, 1)]


def test_elements_repeated_sequence_2():
    para = Paragraph("aaa")
    x = xpath_compile("xxx")
    result = para.elements_repeated_sequence(x, "b")
    assert result == []


def test_get_element_list(sample):
    content_part = sample.content
    elements = content_part.get_elements("//text:p")
    # The annotation paragraph is counted
    assert len(elements) == 8


def test_get_tagname(sample):
    element = sample.para
    assert element.tag == "text:p"


def test_get_parent():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    child = element.get_element("//text:span")
    parent = child.parent
    assert parent.tag == "text:p"


def test_get_root():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    root = element.root
    assert root.parent is None


def test_clone(sample):
    element = sample.para
    copy = element.clone
    assert id(element) != id(copy)
    assert element.text == copy.text


def test_delete_child():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    child = element.get_element("//text:span")
    element.delete(child)
    assert element.serialize() == "<text:p/>"


def test_delete_self():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    child = element.get_element("//text:span")
    child.delete()
    assert element.serialize() == "<text:p/>"


def test_delete_self_2():
    element = Element.from_tag("<text:p><text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    assert element.serialize() == "<text:p>keep</text:p>"


def test_delete_self_3():
    element = Element.from_tag("<text:p>before<text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    assert element.serialize() == "<text:p>beforekeep</text:p>"


def test_delete_self_4():
    element = Element.from_tag("<text:p><tag>x</tag>before<text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    assert element.serialize() == "<text:p><tag>x</tag>beforekeep</text:p>"


def test_delete_self_5():
    element = Element.from_tag("<text:p><tag>x</tag><text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    assert element.serialize() == "<text:p><tag>x</tag>keep</text:p>"


def test_delete_root():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    root = element.root
    with pytest.raises(ValueError):
        root.delete()


def test_get_attributes(sample):
    element = sample.para
    attributes = element.attributes
    excepted = {"text:style-name": "Text_20_body"}
    assert attributes == excepted


def test_get_attribute(sample):
    element = sample.para
    unknown = element.get_attribute("style-name")
    assert unknown is None


def test_get_attribute_namespace(sample):
    element = sample.para
    text = element.get_attribute("text:style-name")
    assert isinstance(text, str)
    assert text == "Text_20_body"


def test_get_attribute_none(sample):
    element = sample.para
    dummy = element.get_attribute("and_now_for_sth_completely_different")
    assert dummy is None


def test_set_attribute(sample):
    element = sample.para
    element.set_attribute("test", "a value")
    assert element.get_attribute("test") == "a value"
    element.del_attribute("test")


def test_set_attribute_namespace(sample):
    element = sample.para
    element.set_attribute("text:style-name", "Note")
    assert element.get_attribute("text:style-name") == "Note"
    element.del_attribute("text:style-name")


def test_set_attribute_special(sample):
    element = sample.para
    element.set_attribute("test", SPECIAL_CHARS)
    assert element.get_attribute("test") == SPECIAL_CHARS
    element.del_attribute("test")


def test_del_attribute(sample):
    element = sample.para
    element.set_attribute("test", "test")
    element.del_attribute("test")
    assert element.get_attribute("test") is None


def test_del_attribute_raise(sample):
    element = sample.para
    with pytest.raises(KeyError):
        element.del_attribute("office:boolean-value")


def test_del_attribute_list(sample):
    element = sample.para
    element.set_attribute("test", "test")
    attr_list = ["text:style-name", "test", "notexist1", "office:boolean-value"]
    element.del_attribute_list(attr_list)
    assert element._canonicalize() == "<text:p>This is the first paragraph.</text:p>"


def test_del_attribute_namespace(sample):
    element = sample.para
    element.set_attribute("text:style-name", "Note")
    element.del_attribute("text:style-name")
    assert element.get_attribute("text:style-name") is None


def test_set_attribute_str_default_1():
    element = Element.from_tag("text:p")
    element._set_attribute_str_default("test", "value", "default_value")
    assert element.serialize() == '<text:p test="value"/>'


def test_set_attribute_str_default_2():
    element = Element.from_tag("text:p")
    element._set_attribute_str_default("test", "default_value", "default_value")
    assert element.serialize() == "<text:p/>"


def test_get_attribute_str_default_1():
    element = Element.from_tag('<text:p test="value"/>')
    result = element._get_attribute_str_default("test", "default_value")
    assert result == "value"


def test_get_attribute_str_default_2():
    element = Element.from_tag("<text:p/>")
    result = element._get_attribute_str_default("test", "default_value")
    assert result == "default_value"


def test_set_attr_transparent_1():
    element = Element.from_tag("<text:p/>")
    element.set_attribute("dr3d:ambient-color", "blue")
    assert element.serialize() == '<text:p dr3d:ambient-color="#0000FF"/>'


def test_set_attr_transparent_2():
    element = Element.from_tag("<text:p/>")
    element.set_attribute("dr3d:ambient-color", "transparent")
    assert element.serialize() == '<text:p dr3d:ambient-color="transparent"/>'


def test_get_text(sample):
    element = sample.para
    text = element.text
    assert text == "This is the first paragraph."


def test_set_text(sample):
    element = sample.para
    old_text = element.text
    new_text = "A test"
    element.text = new_text
    assert element.text == new_text
    element.text = old_text
    assert element.text == old_text


def test_set_text_special(sample):
    element = sample.para
    old_text = element.text
    element.text = SPECIAL_CHARS
    assert element.text == SPECIAL_CHARS
    element.text = old_text


def test_get_text_content(sample):
    element = sample.anno
    text = element.text_content
    print(element.serialize())
    print(repr(text))
    expected = "This is an annotation.\nWith diacritical signs: éè"
    assert text == expected


def test_set_text_content(sample):
    element = sample.anno
    old_text = element.text_content
    text = "Have a break"
    element.text_content = text
    assert element.text_content == text
    element.text_content = old_text
    assert element.text_content == old_text


def test_get_parent2(sample):
    paragraph = sample.para
    parent = paragraph.parent
    assert parent.tag == "text:section"


def test_get_parent_root(sample):
    content = sample.content
    root = content.root
    parent = root.parent
    assert parent is None


def test_insert_element_first_child():
    element = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    child = Element.from_tag("<text:h/>")
    element.insert(child, FIRST_CHILD)
    expected = "<office:text><text:h/><text:p/><text:p/></office:text>"
    assert element.serialize() == expected


def test_insert_element_last_child():
    element = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    child = Element.from_tag("<text:h/>")
    expected = "<office:text><text:p/><text:p/><text:h/></office:text>"
    element.append(child)
    assert element.serialize() == expected


def test_insert_element_next_sibling():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    element = root.get_elements("//text:p")[0]
    sibling = Element.from_tag("<text:h/>")
    expected = "<office:text><text:p/><text:h/><text:p/></office:text>"
    element.insert(sibling, NEXT_SIBLING)
    assert root.serialize() == expected


def test_insert_element_next_sibling_xpath_query():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    xpath_query = xpath_compile("//text:p")
    element = root.get_elements(xpath_query)[0]
    sibling = Element.from_tag("<text:h/>")
    expected = "<office:text><text:p/><text:h/><text:p/></office:text>"
    element.insert(sibling, NEXT_SIBLING)
    assert root.serialize() == expected


def test_insert_element_prev_sibling():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    element = root.get_elements("//text:p")[0]
    sibling = Element.from_tag("<text:h/>")
    element.insert(sibling, PREV_SIBLING)
    assert root.serialize() == "<office:text><text:h/><text:p/><text:p/></office:text>"


def test_insert_element_start_1():
    root = Element.from_tag(
        "<office:text><text:p>abc</text:p><text:p>def</text:p></office:text>"
    )
    element = root.get_elements("//text:p")[0]
    item = Element.from_tag("<text:span>new</text:span>")
    element.insert(item, start=True)
    print(root.serialize())
    assert root.serialize() == (
        "<office:text>"
        "<text:p><text:span>new</text:span>abc</text:p>"
        "<text:p>def</text:p>"
        "</office:text>"
    )


def test_insert_element_start_2():
    root = Element.from_tag(
        "<office:text><text:p>abc</text:p><text:p>def</text:p></office:text>"
    )
    element = root.get_elements("//text:p")[0]
    item = Element.from_tag("<text:span>new</text:span>tail")
    element.insert(item, start=True)
    print(root.serialize())
    assert root.serialize() == (
        "<office:text>"
        "<text:p><text:span>new</text:span>tailabc</text:p>"
        "<text:p>def</text:p>"
        "</office:text>"
    )


def test_insert_element_bad_element():
    element = Element.from_tag("text:p")
    with pytest.raises(AttributeError):
        element.insert("<text:span/>", FIRST_CHILD)


def test_insert_element_bad_position():
    element = Element.from_tag("text:p")
    child = Element.from_tag("text:span")
    with pytest.raises(ValueError):
        element.insert(child, 999)


def test_insert_element_bad_position_2():
    element = Element.from_tag("text:p")
    child = Element.from_tag("text:span")
    with pytest.raises(ValueError):
        element._insert(child, before="a", after="b")


def test_children(sample):
    element = sample.anno
    children = element.children
    assert len(children) == 4
    child = children[0]
    assert child.tag == "dc:creator"
    child = children[-1]
    assert child.tag == "text:p"


def test_get_element_idx():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    xpath_query = xpath_compile("//text:p")
    element = root._get_element_idx(xpath_query, 1)
    assert element.tag == "text:p"


def test_get_element_idx_none():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    xpath_query = xpath_compile("//text:p")
    element = root._get_element_idx(xpath_query, 42)
    assert element is None


def test_get_attribute_integer():
    element = Element.from_tag('<text:s text:c="4" />')
    result = element.get_attribute_integer("text:c")
    assert result == 4


def test_get_attribute_integer_bad():
    element = Element.from_tag('<text:s text:c="oops" />')
    result = element.get_attribute_integer("text:c")
    assert result is None


def test_get_attribute_integer_bad_2():
    element = Element.from_tag("<text:s/>")
    result = element.get_attribute_integer("text:c")
    assert result is None


def test_set_attribute_wrong_color():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(TypeError):
        element.set_attribute("dr3d:ambient-color", False)


def test_get_attribute_integer_bad_3():
    element = Element.from_tag("<text:s />")
    result = element.get_attribute_integer("oops")
    assert result is None


def test_set_text_bad():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(TypeError):
        element.text = False


def test_append_element():
    element = Element.from_tag("text:p")
    element.append("f")
    element.append("oo1")
    element.append(Element.from_tag("text:line-break"))
    element.append("f")
    element.append("oo2")
    expected = "<text:p>foo1<text:line-break/>foo2</text:p>"
    assert element.serialize() == expected


def test_append_wrong_item():
    element = Element.from_tag("text:p")
    with pytest.raises(TypeError):
        element._Element__append([])


def test_search_paragraph(span_styles):
    """Search text in a paragraph."""
    pos = span_styles.para.search("ère")
    expected = 4
    assert pos == expected


def test_match_span_search(span_styles):
    """Search text in a span."""
    pos = span_styles.span.search("moust")
    assert pos == 0


def test_match_inner_span_search(span_styles):
    """Search text in a span from the parent paragraph."""
    pos = span_styles.para.search("roug")
    assert pos == 29


def test_simple_regex_search(span_styles):
    """Search a simple regex."""
    pos = span_styles.para.search("che roug")
    assert pos == 25


def test_intermediate_regex_search(span_styles):
    """Search an intermediate regex."""
    pos = span_styles.para.search("moustache (blanche|rouge)")
    assert pos == 19


def test_complex_regex_search(span_styles):
    """Search a complex regex."""
    # The (?<=...) part is pointless as we don't try to get groups from
    # a MatchObject. However, it's a valid regex expression.
    pos = span_styles.para.search(r"(?<=m)(ou)\w+(che) (blan\2|r\1ge)")
    assert pos == 20


def test_compiled_regex_search(span_styles):
    """Search with a compiled pattern."""
    pattern = re.compile(r"moustache")
    pos = span_styles.para.search(pattern)
    assert pos == 19


def test_failing_match_search(span_styles):
    """Test a regex that doesn't match."""
    pos = span_styles.para.search("Le Père moustache")
    assert pos is None


def test_match_paragraph(span_styles):
    """Match text in a paragraph."""
    match = span_styles.para.match("ère")
    assert match


def test_match_span(span_styles):
    """Match text in a span."""
    match = span_styles.span.match("moust")
    assert match


def test_match_inner_span(span_styles):
    """Match text in a span from the parent paragraph."""
    match = span_styles.para.match("roug")
    assert match


def test_simple_regex(span_styles):
    """Match a simple regex."""
    match = span_styles.para.match("che roug")
    assert match


def test_intermediate_regex(span_styles):
    """Match an intermediate regex."""
    match = span_styles.para.match("moustache (blanche|rouge)")
    assert match


def test_complex_regex(span_styles):
    """Match a complex regex."""
    # The (?<=...) part is pointless as we don't try to get groups from
    # a MatchObject. However, it's a valid regex expression.
    match = span_styles.para.match(r"(?<=m)(ou)\w+(che) (blan\2|r\1ge)")
    assert match


def test_compiled_regex(span_styles):
    """Match with a compiled pattern."""
    pattern = re.compile(r"moustache")
    match = span_styles.para.match(pattern)
    assert match


def test_failing_match(span_styles):
    """Test a regex that doesn't match."""
    match = span_styles.para.match("Le Père moustache")
    assert not match


def test_count(span_styles):
    paragraph = span_styles.para
    expected = paragraph.serialize()
    count = paragraph.replace("ou")
    assert count == 2
    # Ensure the orignal was not altered
    assert paragraph.serialize() == expected


def test_replace(span_styles):
    paragraph = span_styles.para
    clone = paragraph.clone
    count = clone.replace("moustache", "barbe")
    assert count == 1
    expected = "Le Père Noël a une barbe rouge."
    assert str(clone) == expected + "\n"
    assert clone.inner_text == expected
    # Ensure the orignal was not altered
    assert clone.serialize() != paragraph.serialize()


def test_replace_bad(span_styles):
    paragraph = span_styles.para
    clone = paragraph.clone
    with pytest.raises(Exception):  # noqa: B017
        clone.replace([], "barbe")


def test_replace_simple():
    paragraph = Element.from_tag("<text:p>abc</text:p>")
    count = paragraph.replace("abc", "xyz")
    assert count == 1


def test_across_span(span_styles):
    paragraph = span_styles.para
    count = paragraph.replace("moustache rouge")
    assert count == 0


def test_missing(span_styles):
    paragraph = span_styles.para
    count = paragraph.replace("barbe")
    assert count == 0


def test_register():
    register_element_class(DummyElement)
    element = Element.from_tag("office:dummy1")
    assert isinstance(element, DummyElement)


def test_unregistered():
    element = Element.from_tag("office:dummy2")
    assert isinstance(element, Element)
    assert not isinstance(element, DummyElement)


def test_is_bound():
    paragraph = Element.from_tag("<text:p>abc</text:p>")
    assert paragraph.is_bound


def test_content_setter_empty():
    element = Element.from_tag("<office:text><text:p>abc</text:p></office:text>")
    element.text_content = None
    assert element.serialize() == "<office:text><text:p></text:p></office:text>"


def test_is_empty_1():
    element = Element.from_tag("<text:span>abc</text:span>def")
    assert element.is_empty() is False


def test_is_empty_2():
    element = Element.from_tag("<text:span>abc</text:span>")
    assert element.is_empty() is False


def test_is_empty_3():
    element = Element.from_tag("<text:span></text:span>def")
    assert element.is_empty() is False


def test_is_empty_4():
    element = Element.from_tag("<text:span></text:span>")
    assert element.is_empty() is True


def test_xpath_str():
    element = Element.from_tag('<text:s text:c="1" />')
    result = element.xpath("//@text:c")
    assert result == ["1"]


def test_xpath_no_bool():
    element = Element.from_tag('<text:s text:c="1" />')
    result = element.xpath("boolean(//@text:c)")
    assert result == []


def test_serialize():
    element = Element.from_tag("<text:span>abc</text:span>")
    assert element.serialize() == "<text:span>abc</text:span>"


def test_serialize_with_ns():
    element = Element.from_tag("<text:span>abc</text:span>")
    assert element.serialize(with_ns=True) == (
        "<text:span xmlns:text="
        '"urn:oasis:names:tc:opendocument:xmlns:text:1.0">'
        "abc</text:span>"
    )


def test_set_inner_text_exists():
    element = Element.from_tag("<text:p>x <text:span>abc</text:span> y</text:p>")
    element._set_inner_text("text:span", "new")
    assert element.serialize() == "<text:p>x <text:span>new</text:span> y</text:p>"


def test_set_inner_text_none():
    element = Element.from_tag("<text:p>x y</text:p>")
    element._set_inner_text("text:span", "new")
    assert element.serialize() == "<text:p>x y<text:span>new</text:span></text:p>"


def test_get_inner_text_content():
    element = Element.from_tag("<text:p>x y<text:span>content</text:span></text:p>")
    result = element._get_inner_text("text:span")
    assert result == "content"


def test_get_inner_text_no_content():
    element = Element.from_tag("<text:p>x y</text:p>")
    result = element._get_inner_text("text:span")
    assert result is None


def test_sections(sample):
    content = sample.content
    result = content.body.sections
    assert len(result) == 2
    assert isinstance(result[0], Section)


@pytest.mark.parametrize(
    "tag",
    [
        "office:chart",
        "office:drawing",
        "office:presentation",
        "office:spreadsheet",
        "office:text",
    ],
)
def test_append_named_range_in_class(tag):
    nr = NamedRange("nr_name", "A1:C2", "table name é", usage="filter")
    element = Element.from_tag(tag)
    element.append_named_range(nr)
    result = element.get_named_range(name="nr_name")
    assert isinstance(result, NamedRange)


def test_element_get_attribute_int_default_1():
    element = Element.from_tag("<form:form/>")
    result = element._get_attribute_int_default("form:tab-index", 0)
    assert result == 0


def test_element_get_attribute_int_default_2():
    element = Element.from_tag("<form:form/>")
    element._set_attribute_int_default("form:tab-index", 1, 0)
    result = element._get_attribute_int_default("form:tab-index", 0)
    assert result == 1


def test_element_get_attribute_int_default_3():
    element = Element.from_tag("<form:form/>")
    element._set_attribute_int_default("form:tab-index", "not_an_int", 0)
    result = element._get_attribute_int_default("form:tab-index", 0)
    assert result == 0



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.table import Table


@pytest.fixture
def hidden_ods(samples) -> Iterable[Document]:
    document = Document(samples("minimal_hidden.ods"))
    yield document


def test_display_default():
    table = Table("A Table")
    expected = '<table:table table:name="A Table"/>'
    assert table.serialize() == expected


def test_table_style_get_table_name():
    doc = Document("spreadsheet")
    body = doc.body
    body.clear()
    table = Table("Table0")
    body.append(table)
    table = doc._get_table("Table0")
    assert table.name == "Table0"


def test_table_style_table_display_default():
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table("Table0"))
    assert doc.get_table_displayed("Table0")


def test_table_style_table_display_default_idx():
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table("Table0"))
    assert doc.get_table_displayed(0)


def test_table_style_table_display_change1():
    name0 = "Table0"
    name1 = "Table1"
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table(name0))
    doc.body.append(Table(name1))
    doc.set_table_displayed(name0, True)
    assert doc.get_table_displayed(name0)
    assert doc.get_table_displayed(name1)


def test_table_style_table_display_change2():
    name0 = "Table0"
    name1 = "Table1"
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table(name0))
    doc.body.append(Table(name1))
    doc.set_table_displayed(name0, False)
    assert not doc.get_table_displayed(name0)
    assert doc.get_table_displayed(name1)


def test_table_style_table_display_change3():
    name0 = "Table0"
    name1 = "Table1"
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table(name0))
    doc.body.append(Table(name1))
    doc.set_table_displayed(name0, True)
    doc.set_table_displayed(name1, False)
    assert doc.get_table_displayed(name0)
    assert not doc.get_table_displayed(name1)


def test_table_style_table_display_change1_idx():
    name0 = "Table0"
    name1 = "Table1"
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table(name0))
    doc.body.append(Table(name1))
    doc.set_table_displayed(0, True)
    assert doc.get_table_displayed(0)
    assert doc.get_table_displayed(1)


def test_table_style_table_display_change2_idx():
    name0 = "Table0"
    name1 = "Table1"
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table(name0))
    doc.body.append(Table(name1))
    doc.set_table_displayed(0, False)
    assert not doc.get_table_displayed(0)
    assert doc.get_table_displayed(1)


def test_table_style_table_display_change3_idx():
    name0 = "Table0"
    name1 = "Table1"
    doc = Document("spreadsheet")
    doc.body.clear()
    doc.body.append(Table(name0))
    doc.body.append(Table(name1))
    doc.set_table_displayed(0, True)
    doc.set_table_displayed(1, False)
    assert doc.get_table_displayed(0)
    assert not doc.get_table_displayed(1)


def test_table_style_get_table_name_file(hidden_ods):
    table = hidden_ods._get_table("Tab 1")
    assert table.name == "Tab 1"


def test_table_style_table_display_file1(hidden_ods):
    assert hidden_ods.get_table_displayed("Tab 1")


def test_table_style_table_display_file2(hidden_ods):
    assert not hidden_ods.get_table_displayed("Tab 2")


def test_table_style_table_display_3_idx_file(hidden_ods):
    hidden_ods.body.append(Table("Table 3"))
    assert hidden_ods.get_table_displayed(2)


def test_table_style_table_display_change1_file(hidden_ods):
    hidden_ods.set_table_displayed("Tab 1", True)
    assert hidden_ods.get_table_displayed("Tab 1")
    assert not hidden_ods.get_table_displayed("Tab 2")


def test_table_style_table_display_change2_file(hidden_ods):
    hidden_ods.set_table_displayed("Tab 1", False)
    assert not hidden_ods.get_table_displayed("Tab 1")
    assert not hidden_ods.get_table_displayed("Tab 2")


def test_table_style_table_display_change3_file(hidden_ods):
    hidden_ods.set_table_displayed("Tab 1", False)
    hidden_ods.set_table_displayed("Tab 2", True)
    assert not hidden_ods.get_table_displayed("Tab 1")
    assert hidden_ods.get_table_displayed("Tab 2")


def test_table_style_table_display_change1_idx_file(hidden_ods):
    hidden_ods.set_table_displayed(0, True)
    assert hidden_ods.get_table_displayed(0)
    assert not hidden_ods.get_table_displayed(1)


def test_table_style_table_display_change2_idx_file(hidden_ods):
    hidden_ods.set_table_displayed(0, False)
    assert not hidden_ods.get_table_displayed(0)
    assert not hidden_ods.get_table_displayed(1)


def test_table_style_table_display_change3_idx_file(hidden_ods):
    hidden_ods.set_table_displayed(0, False)
    hidden_ods.set_table_displayed(1, True)
    assert not hidden_ods.get_table_displayed(0)
    assert hidden_ods.get_table_displayed(1)



#!/usr/bin/env python
"""Update the table of contents of a document."""

import os
import sys
from pathlib import Path

from odfdo import Document, Header, Paragraph

_DOC_SEQUENCE = 38
DATA = Path(__file__).parent / "data"
SOURCE = "doc_with_toc.odt"
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "modified_toc"
TARGET = "document.odt"


def read_source_document() -> Document:
    """Return the source Document."""
    try:
        source = sys.argv[1]
    except IndexError:
        source = DATA / SOURCE
    return Document(source)


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def update_toc(document: Document) -> None:
    """Make updates and changes to the document's table of contents."""
    check_toc_v1(document)
    add_some_header(document)
    check_toc_v2(document)
    change_toc_title(document)
    check_toc_v3(document)
    change_toc_title_to_empty(document)
    check_toc_v4(document)
    remove_second_header_1b(document)
    check_toc_v5(document)
    add_toc_title(document)
    check_toc_v6(document)


def check_toc_v1(document: Document) -> None:
    toc = document.body.toc
    content = str(toc).split("\n")
    assert len(content) == 5
    assert content[0].startswith("Table of Contents")
    assert content[1].startswith("1. Lorem 1")
    assert content[2].startswith("1.1. Lorem 1A")
    assert content[3].startswith("1.2. Lorem 1B")
    assert content[4].startswith("1.3. Lorem 1C")


def add_some_header(document: Document) -> None:
    header = Header(1, "New header")
    document.body.append(header)
    document.body.append(Paragraph("Some text after the new header."))
    # update the table of contents
    toc = document.body.toc
    toc.fill(document)


def check_toc_v2(document: Document) -> None:
    toc = document.body.toc
    content = str(toc).split("\n")
    assert len(content) == 6
    assert content[0].startswith("Table of Contents")
    assert content[1].startswith("1. Lorem 1")
    assert content[2].startswith("1.1. Lorem 1A")
    assert content[3].startswith("1.2. Lorem 1B")
    assert content[4].startswith("1.3. Lorem 1C")
    assert content[5].startswith("2. New header")


def change_toc_title(document: Document) -> None:
    toc = document.body.toc
    toc.set_toc_title("Another title")
    toc.fill(document)


def check_toc_v3(document: Document) -> None:
    toc = document.body.toc
    content = str(toc).split("\n")
    assert len(content) == 6
    assert content[0].startswith("Another title")


def change_toc_title_to_empty(document: Document) -> None:
    toc = document.body.toc
    toc.set_toc_title("")  # that will remove the title
    toc.fill(document)


def check_toc_v4(document: Document) -> None:
    toc = document.body.toc
    content = str(toc).split("\n")
    assert len(content) == 5
    assert content[0].startswith("1. Lorem 1")
    assert content[1].startswith("1.1. Lorem 1A")
    assert content[2].startswith("1.2. Lorem 1B")
    assert content[3].startswith("1.3. Lorem 1C")
    assert content[4].startswith("2. New header")


def remove_second_header_1b(document: Document) -> None:
    # find second header:
    header = document.body.get_header(position=2)
    # this 'header' variable is attached to the document, so
    # deleting will remove the element from the document
    header.delete()

    toc = document.body.toc
    toc.fill(document)


def check_toc_v5(document: Document) -> None:
    toc = document.body.toc
    content = str(toc).split("\n")
    assert len(content) == 4
    assert content[0].startswith("1. Lorem 1")
    assert content[1].startswith("1.1. Lorem 1A")
    assert content[2].startswith("1.2. Lorem 1C")
    assert content[3].startswith("2. New header")


def add_toc_title(document: Document) -> None:
    toc = document.body.toc
    toc.set_toc_title("A new title")
    toc.fill(document)


def check_toc_v6(document: Document) -> None:
    toc = document.body.toc
    content = str(toc).split("\n")
    assert len(content) == 5
    assert content[0].startswith("A new title")
    assert content[1].startswith("1. Lorem 1")
    assert content[2].startswith("1.1. Lorem 1A")
    assert content[3].startswith("1.2. Lorem 1C")
    assert content[4].startswith("2. New header")


def main() -> None:
    document = read_source_document()
    update_toc(document)
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    check_toc_v6(document)


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.link import Link


@pytest.fixture
def sample_body(samples) -> Iterable[Element]:
    document = Document(samples("base_text.odt"))
    yield document.body


def test_create_link1():
    link = Link("http://example.com/")
    expected = '<text:a xlink:href="http://example.com/"/>'
    assert link.serialize() == expected


def test_create_link2():
    link = Link(
        "http://example.com/",
        name="link2",
        target_frame="_blank",
        style="style1",
        visited_style="style2",
    )
    expected = (
        '<text:a xlink:href="http://example.com/" '
        'office:name="link2" office:target-frame-name="_blank" '
        'xlink:show="new" text:style-name="style1" '
        'text:visited-style-name="style2"/>'
    )
    assert link.serialize() == expected


def test_get_link(sample_body):
    link1 = Link("http://example.com/", name="link1")
    link2 = Link("http://example.com/", name="link2")
    paragraph = sample_body.get_paragraph()
    paragraph.append(link1)
    paragraph.append(link2)
    element = sample_body.get_link(name="link2")
    expected = '<text:a xlink:href="http://example.com/" office:name="link2"/>'
    assert element.serialize() == expected


def test_get_link_list(sample_body):
    link1 = Link("http://example.com/", name="link1")
    link2 = Link("http://example.com/", name="link2")
    paragraph = sample_body.get_paragraph()
    paragraph.append(link1)
    paragraph.append(link2)
    element = sample_body.get_links()[1]
    expected = '<text:a xlink:href="http://example.com/" office:name="link2"/>'
    assert element.serialize() == expected


def test_get_link_list_name(sample_body):
    link1 = Link("http://example.com/", name="link1", title="title1")
    link2 = Link("http://example.com/", name="link2", title="title2")
    paragraph = sample_body.get_paragraph()
    paragraph.append(link1)
    paragraph.append(link2)
    # name
    element = sample_body.get_links(name="link1")[0]
    expected = (
        '<text:a xlink:href="http://example.com/" '
        'office:name="link1" office:title="title1"/>'
    )
    assert element.serialize() == expected


def test_get_link_list_title(sample_body):
    link1 = Link("http://example.com/", name="link1", title="title1")
    link2 = Link("http://example.com/", name="link2", title="title2")
    paragraph = sample_body.get_paragraph()
    paragraph.append(link1)
    paragraph.append(link2)
    # title
    element = sample_body.get_links(title="title2")[0]
    expected = (
        '<text:a xlink:href="http://example.com/" '
        'office:name="link2" office:title="title2"/>'
    )
    assert element.serialize() == expected


def test_get_link_list_href(sample_body):
    link1 = Link("http://example.com/", name="link1", title="title1")
    link2 = Link("http://example.com/", name="link2", title="title2")
    paragraph = sample_body.get_paragraph()
    paragraph.append(link1)
    paragraph.append(link2)
    # url
    elements = sample_body.get_links(url=r"\.com")
    assert len(elements) == 3


def test_href_from_existing_document(sample_body):
    links = sample_body.get_links(url=r"odfdo")
    assert len(links) == 1


def test_get_link_list_name_and_title(sample_body):
    link1 = Link("http://example.com/", name="link1", title="title1")
    link2 = Link("http://example.com/", name="link2", title="title2")
    paragraph = sample_body.get_paragraph()
    paragraph.append(link1)
    paragraph.append(link2)
    # name and title
    element = sample_body.get_links(name="link1", title="title1")[0]
    expected = (
        '<text:a xlink:href="http://example.com/" '
        'office:name="link1" office:title="title1"/>'
    )
    assert element.serialize() == expected


def test_get_link_by_href(sample_body):
    link = sample_body.get_link(url=r"odfdo")
    url = link.get_attribute("xlink:href")
    assert url == "https://github.com/jdum/odfdo"


def test_get_link_by_path_context(sample_body):
    section2 = sample_body.get_section(position=1)
    link = section2.get_link(url=r"github")
    url = link.url
    assert url == "https://github.com/jdum/odfdo"


def test_get_link_list_not_found(sample_body):
    link1 = Link("http://example.com/", name="link1", title="title1")
    link2 = Link("http://example.com/", name="link2", title="title2")
    paragraph = sample_body.get_paragraph()
    paragraph.append(link1)
    paragraph.append(link2)
    # Not found
    element = sample_body.get_links(name="link1", title="title2")
    assert element == []


def test_insert_link_simple():
    paragraph = Element.from_tag("<text:p>toto tata titi</text:p>")
    paragraph.set_link("http://example.com", regex="tata")
    expected = (
        "<text:p>toto "
        '<text:a xlink:href="http://example.com">tata</text:a> '
        "titi</text:p>"
    )
    assert paragraph.serialize() == expected


def test_insert_link_medium():
    paragraph = Element.from_tag(
        "<text:p><text:span>toto</text:span> tata titi</text:p>"
    )
    paragraph.set_link("http://example.com", regex="tata")
    expected = (
        "<text:p><text:span>toto</text:span> "
        '<text:a xlink:href="http://example.com">tata</text:a> '
        "titi</text:p>"
    )
    assert paragraph.serialize() == expected


def test_insert_link_complex():
    paragraph = Element.from_tag(
        "<text:p>toto <text:span> tata </text:span> titi</text:p>"
    )
    paragraph.set_link("http://example.com", regex="tata")
    expected = (
        "<text:p>toto <text:span> "
        '<text:a xlink:href="http://example.com">'
        "tata</text:a> </text:span> titi"
        "</text:p>"
    )
    assert paragraph.serialize() == expected


def test_repr(sample_body):
    link = sample_body.get_link(url=r"odfdo")
    assert repr(link) == "<Link tag=text:a link=https://github.com/jdum/odfdo>"


def test_str(sample_body):
    link = sample_body.get_link(url=r"odfdo")
    assert str(link) == "[an external link](https://github.com/jdum/odfdo)"


def test_str2(sample_body):
    link = Link("https://example.com/")
    assert str(link) == "(https://example.com/)"


def test_target_frame_1(sample_body):
    link = Link("https://example.com/", target_frame="_blank")
    assert str(link) == "(https://example.com/)"
    assert link.target_frame == "_blank"
    assert link.show == "new"


def test_target_frame_2(sample_body):
    link = Link("https://example.com/", target_frame="replace")
    assert str(link) == "(https://example.com/)"
    assert link.target_frame == "replace"
    assert link.show == "replace"



# Copyright 2018-2026 Jérôme Dumonteil
# Authors (odfdo project): jerome.dumonteil@gmail.com
from __future__ import annotations

import io
import subprocess
import sys
from pathlib import Path

import pytest

from odfdo.document import Document
from odfdo.scripts import highlight
from odfdo.scripts.highlight import main as main_script
from odfdo.scripts.highlight import main_highlight, parse_cli_args

SCRIPT = Path(highlight.__file__)


def run_params(params: list):
    command = [sys.executable, SCRIPT] + params
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_highlight_no_param():
    params = []
    out, err, exitcode = run_params(params)
    print(out, err, exitcode)
    assert exitcode == 2
    assert b"usage:" in err
    assert b"required" in err
    assert b"pattern" in err


# direct access to internal function


def test_highlight_2_no_param_on_main_function(monkeypatch, capsys):
    with pytest.raises(SystemExit) as result:
        monkeypatch.setattr(sys, "argv", [])
        main_script()
        assert result.value.code >= 1
    captured = capsys.readouterr()

    assert "usage:" in captured.err
    assert "required" in captured.err
    assert "pattern" in captured.err


def test_highlight_2_no_param(monkeypatch):
    with pytest.raises(SystemExit) as result:
        params = parse_cli_args([])
        main_script(params)
        assert result.value.code >= 1


def test_highlight_2_some_param_on_main_function(monkeypatch, capsysbinary, samples):
    source = str(samples("base_text.odt"))

    monkeypatch.setattr(sys, "argv", ["odfdo-highlight", "-i", source, "-b", "odfdo"])
    main_script()
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    display_name = " ".join(["odfdo", "highlight", "bold"])
    name = display_name.replace(" ", "_20_")
    assert len(document.body.get_spans(style=name)) == 1


def test_highlight_2_version(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--version"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "odfdo-highlight v3" in captured.out


def test_highlight_2_help(capsys):
    with pytest.raises(SystemExit) as result:
        parse_cli_args(["--help"])
        assert result.value.code == 0
    captured = capsys.readouterr()

    assert "Search for a regular expression" in captured.out


def test_highlight_2_no_style():
    params = parse_cli_args(["pattern"])

    with pytest.raises(SystemExit) as result:
        main_highlight(params)
        assert result.value.code >= 1


def test_highlight_2_no_file():
    params = parse_cli_args(["-i", "none_file1", "-o", "none_file2", "-b", "pattern"])

    with pytest.raises(FileNotFoundError) as result:
        main_highlight(params)
        assert result.value.code >= 1


def test_highlight_2_replace1(capsysbinary, samples):
    source = str(samples("base_text.odt"))
    params = parse_cli_args(["-i", source, "-b", "odfdo"])

    main_highlight(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    display_name = " ".join(["odfdo", "highlight", "bold"])
    name = display_name.replace(" ", "_20_")
    assert len(document.body.get_spans(style=name)) == 1


def test_highlight_2_replace2(capsysbinary, samples):
    source = str(samples("base_text.odt"))
    params = parse_cli_args(["-i", source, "-a", "-b", "This"])

    main_highlight(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    display_name = " ".join(["odfdo", "highlight", "italic", "bold"])
    name = display_name.replace(" ", "_20_")
    assert len(document.body.get_spans(style=name)) == 6


def test_highlight_2_replace3(capsysbinary, samples):
    source = str(samples("base_text.odt"))
    params = parse_cli_args(["-i", source, "-c", "red", "paragraph"])

    main_highlight(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    display_name = " ".join(["odfdo", "highlight", "red"])
    name = display_name.replace(" ", "_20_")
    assert len(document.body.get_spans(style=name)) == 7


def test_highlight_2_replace4(capsysbinary, samples):
    source = str(samples("base_text.odt"))
    params = parse_cli_args(["-i", source, "-c", "red", "-g", "yellow", "paragraph"])

    main_highlight(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    display_name = " ".join(["odfdo", "highlight", "red", "yellow"])
    name = display_name.replace(" ", "_20_")
    assert len(document.body.get_spans(style=name)) == 7


def test_highlight_2_replace4_duplicate(capsysbinary, tmp_path, samples):
    source = str(samples("base_text.odt"))
    dest = tmp_path / "replaced.odt"
    params = parse_cli_args(
        [
            "-i",
            source,
            "-c",
            "red",
            "-g",
            "yellow",
            "-o",
            str(dest),
            "paragraph",
        ]
    )

    main_highlight(params)

    params2 = parse_cli_args(
        [
            "-i",
            str(dest),
            "-c",
            "red",
            "-g",
            "yellow",
            "paragraph",
        ]
    )

    main_highlight(params2)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    display_name = " ".join(["odfdo", "highlight", "red", "yellow"])
    name = display_name.replace(" ", "_20_")
    assert len(document.body.get_spans(style=name)) == 14


def test_highlight_2_toc_replaced(capsysbinary, samples):
    source = str(samples("toc_done.odt"))
    params = parse_cli_args(["-i", source, "-c", "red", "-g", "yellow", "title"])

    main_highlight(params)
    captured = capsysbinary.readouterr()

    content = io.BytesIO(captured.out)
    document = Document(content)
    content.close()
    display_name = " ".join(["odfdo", "highlight", "red", "yellow"])
    name = display_name.replace(" ", "_20_")
    assert len(document.body.get_spans(style=name)) == 12



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

import pytest

from odfdo.utils import bytes_to_str, str_to_bytes, to_bytes, to_str


def test_to_bytes_1():
    assert to_bytes("abc") == b"abc"


def test_to_bytes_2():
    assert to_bytes(b"abc") == b"abc"


def test_to_bytes_3():
    assert to_bytes(123) == 123


def test_to_str_1():
    assert to_str("abc") == "abc"


def test_to_str_2():
    assert to_str(b"abc") == "abc"


def test_to_str_3():
    assert to_str(123) == 123


def test_str_to_bytes_1():
    assert str_to_bytes("abc") == b"abc"


def test_str_to_bytes_2():
    with pytest.raises(AttributeError):
        str_to_bytes(b"abc")


def test_str_to_bytes_3():
    with pytest.raises(AttributeError):
        str_to_bytes(123)


def test_bytes_to_str_1():
    with pytest.raises(AttributeError):
        bytes_to_str("abc")


def test_bytes_to_str_2():
    assert bytes_to_str(b"abc") == "abc"


def test_bytes_to_str_3():
    with pytest.raises(AttributeError):
        bytes_to_str(123)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python


from odfdo.cell import Cell
from odfdo.document import Document
from odfdo.frame import DrawTextBox, Frame
from odfdo.header import Header
from odfdo.line_break import LineBreak
from odfdo.link import Link
from odfdo.list import List, ListItem
from odfdo.mixin_md import (
    MD_GLOBAL,
    LIStyle,
    SplitSpace,
    _as_bold,
    _as_bold_italic,
    _as_none,
    _as_strike,
    _copy_global,
    _md_swap_spaces,
    _restore_global,
    _set_global,
)
from odfdo.note import Note
from odfdo.paragraph import Paragraph, Span
from odfdo.table import Table
from odfdo.toc import TOC
from odfdo.tracked_changes import TrackedChanges


def test_md_swap_spaces_empty():
    word = ""
    assert _md_swap_spaces(word) == SplitSpace("", "", "")


def test_set_global():
    doc = "any"
    _set_global(doc)
    assert MD_GLOBAL["document"] == doc
    assert MD_GLOBAL["list_level"] == {}
    assert MD_GLOBAL["footnote"] == []
    assert MD_GLOBAL["endnote"] == []


def test_copy_global():
    doc = "any"
    _set_global(doc)
    glob_copy = _copy_global()
    assert glob_copy["document"] == "any"
    _set_global("other")
    assert glob_copy["document"] == "any"


def test_restore_global():
    data = {"document": "initial"}
    doc = "any"
    _set_global(doc)
    _restore_global(data)
    assert MD_GLOBAL["document"] == "initial"


def test_as_bold_empty():
    text = " "
    assert _as_bold(text) == text


def test_as_bold_italic_empty():
    text = " "
    assert _as_bold_italic(text) == text


def test_as_strike_empty():
    text = " "
    assert _as_strike(text) == text


def test_md_style_no_doc_no_style():
    p = Paragraph()
    _set_global(None)
    assert p._md_is_fixed_paragraph() is False


def test_md_style_no_doc():
    p = Paragraph()
    p.style = "some_style"
    _set_global(None)
    assert p._md_is_fixed_paragraph() is False


def test_md_style_no_style():
    doc = Document()
    _set_global(doc)
    p = Paragraph()
    p.style = "unknown_style"
    assert p._md_is_fixed_paragraph() is False


def test_md_style_styling_1():
    doc = Document()
    _set_global(doc)
    p = Paragraph()
    p.style = "unknown_style"
    result = p._md_styling()
    assert result == _as_none


def test_md_style_styling_no_doc():
    _set_global(None)
    p = Paragraph()
    p.style = "unknown_style"
    result = p._md_styling()
    assert result == _as_none


def test_md_toc_1():
    toc = TOC()
    index_body = toc.get_element("text:index-body")
    index_body.delete()
    result = toc._md_format()
    assert result == ""


def test_md_toc_2():
    toc = TOC()
    index_body = toc.get_element("text:index-body")
    index_body.delete()
    result = toc._md_collect()
    assert result == []


def test_md_note():
    _set_global(None)
    note = Note()
    result = note._md_collect()
    assert result == ["[]\n"]


def test_md_zap():
    tc = TrackedChanges()
    result = tc._md_format()
    assert result == ""


def test_md_lb():
    lb = LineBreak()
    result = lb._md_collect()
    assert result == ["\\\n"]


def test_md_paragraph_tail():
    p = Paragraph("text")
    p.tail = "but paragraph have no tail"
    result = p._md_collect_fixed_text()
    assert result == "```\ntext\n```\nbut paragraph have no tail"


def test_md_paragraph_listyle_no_doc():
    _set_global(None)
    p = Paragraph()
    p.style = "unknown_style"
    result = p._md_collect_list_item_style()
    assert result == LIStyle("", "")


def test_md_paragraph_listyle_no_style():
    doc = Document()
    _set_global(doc)
    p = Paragraph()
    p.style = "unknown_style"
    result = p._md_collect_list_item_style()
    assert result == LIStyle("", "")


def test_md_paragraph_listyle_no_list_style():
    doc = Document()
    _set_global(doc)
    p = Paragraph()
    p.style = "Standard"
    result = p._md_collect_list_item_style()
    assert result == LIStyle("", "")


def test_md_header_level0():
    h = Header(0, "title")
    result = h._md_format()
    assert result == "title"


def test_md_list_item():
    item = ListItem(Header(1, "B"))
    result = item._md_format()
    assert result == " -  # B"


def test_md_list_buggy():
    li = List(["A", "B"])
    li.insert(Paragraph("this never happens"), position=1)
    result = li._md_format()
    assert result == " -  A\n -  B"


def test_md_list_empty():
    li = List()
    result = li._md_format()
    assert result == ""


def test_md_list_empty_collect():
    li = List()
    result = li._md_collect()
    assert result == []


def test_md_span_empty_collect():
    s = Span()
    result = s._md_collect()
    assert result == []


def test_md_span_text_collect():
    s = Span("text")
    result = s._md_collect()
    assert result == ["text"]


def test_md_link_text():
    url = Link("http://example.com", text="example")
    result = url._md_collect()
    assert result == ["[example](http://example.com)"]


def test_md_link_no_text():
    url = Link("http://example.com")
    result = url._md_collect()
    assert result == ["(http://example.com)"]


def test_md_link_empty():
    url = Link()
    result = url._md_collect()
    assert result == ["()"]


def test_md_draw_text_box_empty():
    dtb = DrawTextBox()
    result = dtb._md_collect()
    assert result == []


def test_md_draw_text_something():
    dtb = DrawTextBox()
    dtb.append(Paragraph("strange"))
    result = dtb._md_collect()
    assert result == ["strange"]


def test_md_frame_empty():
    frame = Frame()
    # frame.append(Paragraph("strange"))
    result = frame._md_collect()
    assert result == []


def test_md_frame_something():
    frame = Frame()
    frame.append(Paragraph("strange"))
    result = frame._md_collect()
    assert result == ["strange"]


def test_md_table_empty_collect():
    table = Table("sheet")
    result = table._md_collect()
    assert result == []


def test_md_table_mini_format_1():
    table = Table("sheet")
    table.set_cell((0, 0), Cell(1))
    result = table._md_format()
    assert result == "| 1 |\n|---|\n"


def test_md_table_mini_format_1_none():
    table = Table("sheet")
    table.set_cell((0, 0), Cell(None))
    result = table._md_format()
    assert result == "|   |\n|---|\n"


def test_md_table_mini_format_1_lg():
    table = Table("sheet")
    table.set_cell((0, 0), Cell("abcdef"))
    result = table._md_format()
    assert result == "| abcdef |\n|--------|\n"


def test_md_table_mini_format_2_2():
    table = Table("sheet")
    table.set_cell((0, 0), Cell(1))
    table.set_cell((1, 0), Cell(2))
    table.set_cell((0, 1), Cell(3))
    table.set_cell((1, 1), Cell(4))
    result = table._md_format()
    assert result == "| 1 | 2 |\n|---|---|\n| 3 | 4 |\n"


def test_md_table_mini_format_3_1():
    table = Table("sheet")
    table.set_cell((0, 0), Cell(1))
    table.set_cell((1, 0), Cell(2))
    table.set_cell((2, 0), Cell(3))
    result = table._md_format()
    assert result == "| 1 | 2 | 3 |\n|---|---|---|\n"


def test_md_table_mini_format_3_1_none_6():
    table = Table("sheet")
    table.set_cell((0, 0), Cell(1))
    table.set_cell((1, 0), Cell(None))
    table.set_cell((2, 0), Cell("abcdef"))
    result = table._md_format()
    assert result == "| 1 |   | abcdef |\n|---|---|--------|\n"


def test_md_table_mini_format_1_3_none_6():
    table = Table("sheet")
    table.set_cell((0, 0), Cell(1))
    table.set_cell((0, 1), Cell(None))
    table.set_cell((0, 2), Cell("abcdef"))
    result = table._md_format()
    assert result == "| 1      |\n|--------|\n|        |\n| abcdef |\n"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.const import FOLDER
from odfdo.document import Document
from odfdo.element import NEXT_SIBLING
from odfdo.frame import Frame
from odfdo.image import DrawFillImage, DrawImage, DrawMarker

IMG_PATH = "Pictures/100002010000012C00000042188DCB81589D2C10.png"


@pytest.fixture
def sample_body(samples) -> Iterable[Element]:
    document = Document(samples("frame_image.odp"))
    yield document.body


def test__draw_fill_image_class():
    dfi = DrawFillImage()
    assert isinstance(dfi, DrawFillImage)


def test_create_image():
    image = DrawImage(IMG_PATH)
    expected = (
        f'<draw:image xlink:href="{IMG_PATH}" xlink:type="simple" '
        'xlink:show="embed" xlink:actuate="onLoad"/>'
    )
    assert image.serialize() == expected


def test_get_image_list(sample_body):
    result = sample_body.get_images()
    assert len(result) == 1
    element = result[0]
    assert element.url == IMG_PATH


def test_get_image_list_property(sample_body):
    result = sample_body.images
    assert len(result) == 1
    element = result[0]
    assert element.url == IMG_PATH


def test_get_image_by_name(sample_body):
    element = sample_body.get_image(name="odfdo")
    assert element.url == IMG_PATH


def test_get_image_by_name_none(sample_body):
    element = sample_body.get_image(name="nope")
    assert element is None


def test_get_image_by_position(sample_body):
    element = sample_body.get_image(position=0)
    assert element.url == IMG_PATH


def test_get_image_by_path(sample_body):
    element = sample_body.get_image(url=".png")
    assert element.url == IMG_PATH


def test_insert_image(tmp_path, samples):
    document = Document(samples("frame_image.odp"))
    body = document.body
    path = "a/path"
    image = DrawImage(path)
    frame = Frame("Image Frame", size=("0cm", "0cm"), style="Graphics")
    frame.append(image)
    body.get_frame().parent.insert(frame, NEXT_SIBLING)
    element = body.get_image(name="Image Frame")
    assert element.url == path
    element = body.get_image(position=1)
    tmp_file = tmp_path / "frame_image.odp"
    document.save(tmp_file, packaging=FOLDER)
    assert element.url == path


def test_repr(sample_body):
    element = sample_body.get_image(name="odfdo")
    assert repr(element) == "<DrawImage tag=draw:image>"


def test_str(sample_body):
    element = sample_body.get_image(name="odfdo")
    assert str(element).endswith(".png)")


def test_draw_marker_class():
    marker = DrawMarker()
    assert marker.serialize() == "<draw:marker/>"


def test_draw_marker_class_tag():
    marker = Element.from_tag("<draw:marker/>")
    assert isinstance(marker, DrawMarker)


def test_draw_marker_xml():
    marker = DrawMarker(
        name="name",
        display_name="display name",
        svg_d="some geometry",
        view_box="some vb",
    )
    assert marker._canonicalize() == (
        "<draw:marker "
        'draw:display-name="display name" '
        'draw:name="name" '
        'svg:d="some geometry" '
        'svg:viewBox="some vb">'
        "</draw:marker>"
    )


def test_draw_marker_name():
    marker = DrawMarker(
        name="name",
        display_name="display name",
        svg_d="some geometry",
        view_box="some vb",
    )
    assert marker.name == "name"


def test_draw_marker_display_name():
    marker = DrawMarker(
        name="name",
        display_name="display name",
        svg_d="some geometry",
        view_box="some vb",
    )
    assert marker.display_name == "display name"


def test_draw_marker_svg_d():
    marker = DrawMarker(
        name="name",
        display_name="display name",
        svg_d="some geometry",
        view_box="some vb",
    )
    assert marker.svg_d == "some geometry"


def test_draw_marker_view_box():
    marker = DrawMarker(
        name="name",
        display_name="display name",
        svg_d="some geometry",
        view_box="some vb",
    )
    assert marker.view_box == "some vb"
